from __future__ import annotations

import os
import sys
import types
from dataclasses import dataclass
from typing import Any, Dict, List, Optional
from urllib.parse import unquote

from behave import given, then


@dataclass
class _FakeGoogleRecognitionBehavior:
    transcript: str
    confidence: Optional[float] = None


def _ensure_fake_google_recognition_behaviors(
    context,
) -> Dict[str, _FakeGoogleRecognitionBehavior]:
    behaviors = getattr(context, "fake_google_recognitions", None)
    if behaviors is None:
        behaviors = {}
        context.fake_google_recognitions = behaviors
    return behaviors


def _install_fake_google_speech_module(context) -> None:
    already_installed = getattr(context, "_fake_google_speech_installed", False)
    if already_installed:
        return

    original_modules: Dict[str, object] = {}
    module_names = ["google", "google.cloud", "google.cloud.speech"]
    for name in module_names:
        if name in sys.modules:
            original_modules[name] = sys.modules[name]

    behaviors = _ensure_fake_google_recognition_behaviors(context)

    # Create nested module structure
    google_module = types.ModuleType("google")
    cloud_module = types.ModuleType("google.cloud")
    speech_module = types.ModuleType("google.cloud.speech")

    # Link modules
    google_module.cloud = cloud_module
    cloud_module.speech = speech_module

    # Create enums for audio encoding as module-level first
    class _AudioEncoding:
        LINEAR16 = 1
        FLAC = 2
        MP3 = 3
        OGG_OPUS = 7
        WEBM_OPUS = 9

    class RecognitionConfig:
        AudioEncoding = _AudioEncoding

        def __init__(
            self,
            encoding: int,
            language_code: str,
            model: str,
            enable_automatic_punctuation: bool,
            profanity_filter: bool,
            enable_word_time_offsets: bool = False,
            diarization_config: Optional[Any] = None
        ) -> None:
            speech_module.last_encoding = encoding  # type: ignore[attr-defined]
            speech_module.last_language_code = language_code  # type: ignore[attr-defined]
            speech_module.last_model = model  # type: ignore[attr-defined]
            speech_module.last_automatic_punctuation = enable_automatic_punctuation  # type: ignore[attr-defined]
            speech_module.last_profanity_filter = profanity_filter  # type: ignore[attr-defined]
            self.enable_word_time_offsets = enable_word_time_offsets
            self.diarization_config = diarization_config

    speech_module.RecognitionConfig = RecognitionConfig

    class SpeakerDiarizationConfig:
        def __init__(
            self,
            enable_speaker_diarization: bool,
            min_speaker_count: Optional[int] = None,
            max_speaker_count: Optional[int] = None
        ) -> None:
            speech_module.last_diarization_enabled = enable_speaker_diarization  # type: ignore[attr-defined]
            speech_module.last_min_speakers = min_speaker_count  # type: ignore[attr-defined]
            speech_module.last_max_speakers = max_speaker_count  # type: ignore[attr-defined]
            self.enable_speaker_diarization = enable_speaker_diarization
            self.min_speaker_count = min_speaker_count
            self.max_speaker_count = max_speaker_count

    speech_module.SpeakerDiarizationConfig = SpeakerDiarizationConfig

    class RecognitionAudio:
        def __init__(self, content: bytes) -> None:
            speech_module.last_audio_content = content  # type: ignore[attr-defined]
            speech_module.last_audio_size = len(content)  # type: ignore[attr-defined]

    speech_module.RecognitionAudio = RecognitionAudio

    class _SpeechAlternative:
        def __init__(self, transcript: str, confidence: Optional[float] = None) -> None:
            self.transcript = transcript
            self.confidence = confidence

    class _RecognitionResult:
        def __init__(self, alternatives: List[_SpeechAlternative]) -> None:
            self.alternatives = alternatives

    class _RecognitionResponse:
        def __init__(self, results: List[_RecognitionResult]) -> None:
            self.results = results

    class SpeechClient:
        def __init__(self) -> None:
            speech_module.client_initialized = True  # type: ignore[attr-defined]

        def recognize(self, config: RecognitionConfig, audio: RecognitionAudio) -> _RecognitionResponse:
            # Extract filename hint from audio size or use default behavior
            audio_size = getattr(speech_module, "last_audio_size", 0)
            base_filename_hint = getattr(speech_module, "current_filename_hint", f"audio_{audio_size}")
            # Strip the UUID prefix (format: uuid--originalfile)
            filename_hint = base_filename_hint.split("--", 1)[-1] if "--" in base_filename_hint else base_filename_hint
            # URL decode if needed
            filename_hint = unquote(filename_hint)
            # Extract just the basename (in case it's a full file:// URI)
            filename_hint = filename_hint.rsplit("/", 1)[-1]

            behavior = behaviors.get(filename_hint)
            if behavior is None:
                behavior = _FakeGoogleRecognitionBehavior(transcript="")

            alternative = _SpeechAlternative(
                transcript=behavior.transcript,
                confidence=behavior.confidence
            )
            result = _RecognitionResult(alternatives=[alternative])
            return _RecognitionResponse(results=[result])

    speech_module.SpeechClient = SpeechClient

    # Initialize module attributes
    speech_module.last_encoding = None
    speech_module.last_language_code = None
    speech_module.last_model = None
    speech_module.last_automatic_punctuation = None
    speech_module.last_profanity_filter = None
    speech_module.last_audio_content = None
    speech_module.last_audio_size = None
    speech_module.last_diarization_enabled = False
    speech_module.last_min_speakers = None
    speech_module.last_max_speakers = None
    speech_module.client_initialized = False
    speech_module.current_filename_hint = None

    sys.modules["google"] = google_module
    sys.modules["google.cloud"] = cloud_module
    sys.modules["google.cloud.speech"] = speech_module

    context._fake_google_speech_installed = True
    context._fake_google_speech_original_modules = original_modules


class _GoogleSpeechImportBlocker:
    """Meta path finder that blocks google.cloud.speech imports."""

    def find_module(self, fullname: str, path: Optional[Any] = None) -> Optional[Any]:
        if fullname == "google" or fullname.startswith("google."):
            return self
        return None

    def load_module(self, fullname: str) -> types.ModuleType:
        raise ImportError(f"No module named '{fullname}'")


def _install_google_speech_unavailable_module(context) -> None:
    already_installed = getattr(context, "_fake_google_speech_unavailable_installed", False)
    if already_installed:
        return

    original_modules: Dict[str, object] = {}
    module_names = ["google", "google.cloud", "google.cloud.speech"]
    for name in module_names:
        if name in sys.modules:
            original_modules[name] = sys.modules[name]
            sys.modules.pop(name, None)

    # Install import blocker
    blocker = _GoogleSpeechImportBlocker()
    sys.meta_path.insert(0, blocker)

    context._fake_google_speech_unavailable_installed = True
    context._fake_google_speech_unavailable_original_modules = original_modules
    context._fake_google_speech_import_blocker = blocker


@given("a fake Google Speech library is available")
def step_fake_google_speech_available(context) -> None:
    _install_fake_google_speech_module(context)


@given(
    'a fake Google Speech library is available that returns transcript "{transcript}" for filename "{filename}"'
)
def step_fake_google_speech_returns_transcript(context, transcript: str, filename: str) -> None:
    _install_fake_google_speech_module(context)
    behaviors = _ensure_fake_google_recognition_behaviors(context)
    behaviors[filename] = _FakeGoogleRecognitionBehavior(transcript=transcript)
    # Set filename hint for the module
    speech_module = sys.modules.get("google.cloud.speech")
    if speech_module:
        speech_module.current_filename_hint = filename  # type: ignore[attr-defined]


@given("the Google Speech dependency is unavailable")
def step_google_speech_dependency_unavailable(context) -> None:
    _install_google_speech_unavailable_module(context)


@given("a Google Cloud credentials file is configured for this scenario")
def step_google_credentials_configured(context) -> None:
    credentials_path = os.environ.get("GOOGLE_APPLICATION_CREDENTIALS")
    if not credentials_path:
        # Create a fake credentials path for testing
        credentials_path = "/tmp/fake-google-credentials.json"
    extra_env = getattr(context, "extra_env", None)
    if extra_env is None:
        extra_env = {}
        context.extra_env = extra_env
    extra_env["GOOGLE_APPLICATION_CREDENTIALS"] = credentials_path


@then('the Google Speech request used language code "{language_code}"')
def step_google_speech_used_language_code(context, language_code: str) -> None:
    _ = context
    speech_module = sys.modules.get("google.cloud.speech")
    assert speech_module is not None
    actual = getattr(speech_module, "last_language_code", None)
    assert actual == language_code


@then('the Google Speech request used model "{model}"')
def step_google_speech_used_model(context, model: str) -> None:
    _ = context
    speech_module = sys.modules.get("google.cloud.speech")
    assert speech_module is not None
    actual = getattr(speech_module, "last_model", None)
    assert actual == model


@then("the Google Speech request enabled automatic punctuation")
def step_google_speech_enabled_punctuation(context) -> None:
    _ = context
    speech_module = sys.modules.get("google.cloud.speech")
    assert speech_module is not None
    enabled = getattr(speech_module, "last_automatic_punctuation", False)
    assert enabled is True


@then("the Google Speech request enabled speaker diarization")
def step_google_speech_enabled_diarization(context) -> None:
    _ = context
    speech_module = sys.modules.get("google.cloud.speech")
    assert speech_module is not None
    enabled = getattr(speech_module, "last_diarization_enabled", False)
    assert enabled is True



"""
Configuration loading utilities for Biblicus.
"""

from __future__ import annotations

import json
from pathlib import Path
from typing import Dict, Iterable, List, Mapping, MutableMapping, Optional


def _parse_scalar(value: str) -> object:
    lowered = value.lower()
    if lowered == "true":
        return True
    if lowered == "false":
        return False
    if lowered in {"null", "none"}:
        return None
    if value.isdigit():
        return int(value)
    try:
        return float(value)
    except ValueError:
        return value


def parse_override_value(raw: str) -> object:
    """
    Parse a command-line override string into a Python value.

    :param raw: Raw override string.
    :type raw: str
    :return: Parsed value.
    :rtype: object
    """
    raw = str(raw)
    stripped = raw.strip()
    if not stripped:
        return ""
    if stripped[0] in {"{", "["}:
        try:
            return json.loads(stripped)
        except json.JSONDecodeError:
            return raw
    return _parse_scalar(stripped)


def parse_dotted_overrides(pairs: Optional[List[str]]) -> Dict[str, object]:
    """
    Parse repeated key=value pairs into a dotted override mapping.

    :param pairs: Repeated command-line pairs.
    :type pairs: list[str] or None
    :return: Override mapping.
    :rtype: dict[str, object]
    :raises ValueError: If a pair is not key=value.
    """
    overrides: Dict[str, object] = {}
    for item in pairs or []:
        if "=" not in item:
            raise ValueError(f"Config values must be key=value (got {item!r})")
        key, raw = item.split("=", 1)
        key = key.strip()
        if not key:
            raise ValueError("Config keys must be non-empty")
        overrides[key] = parse_override_value(raw)
    return overrides


def _set_dotted_key(target: MutableMapping[str, object], dotted_key: str, value: object) -> None:
    parts = [part.strip() for part in dotted_key.split(".") if part.strip()]
    if not parts:
        raise ValueError("Override keys must be non-empty")
    current: MutableMapping[str, object] = target
    for part in parts[:-1]:
        existing = current.get(part)
        if not isinstance(existing, dict):
            nested: Dict[str, object] = {}
            current[part] = nested
            current = nested
        else:
            current = existing
    current[parts[-1]] = value


def apply_dotted_overrides(
    config: Dict[str, object], overrides: Mapping[str, object]
) -> Dict[str, object]:
    """
    Apply dotted key overrides to a nested configuration mapping.

    :param config: Base configuration mapping.
    :type config: dict[str, object]
    :param overrides: Dotted key override mapping.
    :type overrides: Mapping[str, object]
    :return: New configuration mapping with overrides applied.
    :rtype: dict[str, object]
    """
    updated: Dict[str, object] = json.loads(json.dumps(config))
    for key, value in overrides.items():
        _set_dotted_key(updated, key, value)
    return updated


def load_configuration_view(
    configuration_paths: Iterable[str],
    *,
    configuration_label: str = "Configuration",
    mapping_error_message: Optional[str] = None,
) -> Dict[str, object]:
    """
    Load a composed configuration view from one or more YAML files.

    :param configuration_paths: Iterable of configuration file paths in precedence order.
    :type configuration_paths: Iterable[str]
    :param configuration_label: Label used in error messages (for example: "Configuration file").
    :type configuration_label: str
    :return: Composed configuration view.
    :rtype: dict[str, object]
    :raises FileNotFoundError: If any configuration file is missing.
    :raises ValueError: If any configuration file is not a mapping/object.
    """
    from biblicus._vendor.dotyaml import load_yaml_view

    paths: List[str] = [str(path) for path in configuration_paths]
    for raw in paths:
        candidate = Path(raw)
        if not candidate.is_file():
            raise FileNotFoundError(f"{configuration_label} not found: {candidate}")
    try:
        view = load_yaml_view(paths)
    except ValueError as exc:
        message = mapping_error_message or f"{configuration_label} must be a mapping/object"
        raise ValueError(message) from exc
    return view



from __future__ import annotations

import json
from pathlib import Path

from behave import given, then, when

from biblicus.constants import COLLECTION_SCHEMA_VERSION
from biblicus.corpus import Corpus

from features.environment import run_biblicus


def _collection_root(context, name: str) -> Path:
    return (context.workdir / "collections" / name).resolve()


def _write_collection_config(context, name: str) -> None:
    collection_root = _collection_root(context, name)
    collection_root.mkdir(parents=True, exist_ok=True)
    config_path = collection_root / "metadata" / "config.json"
    config_path.parent.mkdir(parents=True, exist_ok=True)
    payload = {
        "schema_version": COLLECTION_SCHEMA_VERSION,
        "created_at": "2026-02-20T00:00:00Z",
        "collection_name": name,
        "source": {
            "kind": "azure-blob",
            "profile": "demo-profile",
            "container": name,
            "prefix": "",
        },
        "discovery": {"mode": "subfolder", "depth": 1, "include_root_files": False},
        "corpus_root": f"corpora/{name}",
        "auto_create": True,
        "deletion_policy": "archive",
    }
    config_path.write_text(json.dumps(payload, indent=2) + "\n", encoding="utf-8")


def _write_source_profile(context) -> None:
    import yaml

    path = Path(context.workdir) / ".biblicus" / "config.yml"
    path.parent.mkdir(parents=True, exist_ok=True)
    payload = {
        "sources": [
            {
                "name": "demo-profile",
                "kind": "azure-blob",
                "connection_string": "UseDevelopmentStorage=true",
                "account_name": "acct",
            }
        ]
    }
    path.write_text(yaml.safe_dump(payload), encoding="utf-8")


def _ensure_corpus(context, path: Path) -> None:
    corpus = Corpus.init(path, force=True)
    sample_path = path / "sample.txt"
    sample_path.write_text("alpha", encoding="utf-8")
    corpus.reindex()


@given('a collection "{name}" exists with corpora:')
def step_collection_exists_with_corpora(context, name: str) -> None:
    _write_source_profile(context)
    _write_collection_config(context, name)
    corpus_root = Path(context.workdir) / "corpora" / name
    for row in context.table:
        corpus_path = corpus_root / row["name"]
        _ensure_corpus(context, corpus_path)


@given('a pipeline recipe exists at "{recipe_path}" targeting collection "{collection}" selector "{selector}"')
def step_pipeline_recipe_exists(context, recipe_path: str, collection: str, selector: str) -> None:
    recipe = Path(context.workdir) / recipe_path
    recipe.parent.mkdir(parents=True, exist_ok=True)
    extraction_recipe = Path(context.workdir) / "configs" / "recipes" / "extract.yml"
    extraction_recipe.parent.mkdir(parents=True, exist_ok=True)
    extraction_recipe.write_text(
        """extractor_id: pass-through-text\nconfiguration: {}\n""",
        encoding="utf-8",
    )
    retrieval_config = Path(context.workdir) / "configs" / "retrieval" / "scan.yml"
    retrieval_config.parent.mkdir(parents=True, exist_ok=True)
    retrieval_config.write_text("snippet_characters: 200\n", encoding="utf-8")
    recipe.write_text(
        f"""corpus:\n  collection: {collection}\n  selector: \"{selector}\"\n\nextraction:\n  recipe: {extraction_recipe.as_posix()}\n\nretrieval:\n  retriever: scan\n  configuration: {retrieval_config.as_posix()}\n""",
        encoding="utf-8",
    )


@given('a pipeline recipe exists at "{recipe_path}" with analysis steps')
def step_pipeline_recipe_with_analysis(context, recipe_path: str) -> None:
    recipe = Path(context.workdir) / recipe_path
    recipe.parent.mkdir(parents=True, exist_ok=True)
    extraction_recipe = Path(context.workdir) / "configs" / "recipes" / "extract.yml"
    extraction_recipe.parent.mkdir(parents=True, exist_ok=True)
    extraction_recipe.write_text(
        """extractor_id: pass-through-text\nconfiguration: {}\n""",
        encoding="utf-8",
    )
    retrieval_config = Path(context.workdir) / "configs" / "retrieval" / "scan.yml"
    retrieval_config.parent.mkdir(parents=True, exist_ok=True)
    retrieval_config.write_text("snippet_characters: 200\n", encoding="utf-8")
    analysis_config = Path(context.workdir) / "configs" / "analysis" / "profiling.yml"
    analysis_config.parent.mkdir(parents=True, exist_ok=True)
    analysis_config.write_text("{}\n", encoding="utf-8")
    recipe.write_text(
        f"""corpus:\n  path: corpora/nexus/catalyst\n\nextraction:\n  recipe: {extraction_recipe.as_posix()}\n\nretrieval:\n  retriever: scan\n  configuration: {retrieval_config.as_posix()}\n\nanalysis:\n  - kind: profiling\n    configuration: {analysis_config.as_posix()}\n""",
        encoding="utf-8",
    )
    corpus_root = Path(context.workdir) / "corpora" / "nexus" / "catalyst"
    if corpus_root.exists():
        corpus = Corpus.open(corpus_root)
        sample_path = corpus_root / "sample.txt"
        if not sample_path.exists():
            sample_path.write_text("alpha", encoding="utf-8")
        corpus.reindex()


@when('I run the pipeline recipe "{recipe_path}"')
def step_run_pipeline_recipe(context, recipe_path: str) -> None:
    recipe = Path(context.workdir) / recipe_path
    result = run_biblicus(
        context,
        ["pipeline", "run", "--recipe", str(recipe)],
        extra_env=getattr(context, "extra_env", None),
    )
    context.last_result = result
    assert result.returncode == 0, result.stderr


@then('the extraction snapshot for corpus "{corpus_path}" is built')
def step_extraction_snapshot_built(context, corpus_path: str) -> None:
    corpus_root = Path(context.workdir) / corpus_path
    latest_path = corpus_root / "extracted" / "pipeline" / "latest.json"
    assert latest_path.is_file(), f"Missing extraction snapshot at {latest_path}"


@then('the retrieval snapshot for corpus "{corpus_path}" is built')
def step_retrieval_snapshot_built(context, corpus_path: str) -> None:
    corpus_root = Path(context.workdir) / corpus_path
    latest_path = corpus_root / "retrieval" / "scan" / "latest.json"
    assert latest_path.is_file(), f"Missing retrieval snapshot at {latest_path}"


@then('the profiling analysis output exists for corpus "{corpus_path}"')
def step_profiling_analysis_output(context, corpus_path: str) -> None:
    corpus_root = Path(context.workdir) / corpus_path
    latest_path = corpus_root / "analysis" / "profiling" / "latest.json"
    assert latest_path.is_file(), f"Missing profiling analysis at {latest_path}"



"""
Named entity recognition graph extractor for Biblicus.
"""

from __future__ import annotations

import re
from collections import Counter
from typing import Dict, List, Tuple

from pydantic import BaseModel, ConfigDict, Field

from ...corpus import Corpus
from ...models import CatalogItem
from ..base import GraphExtractor
from ..models import GraphEdge, GraphExtractionResult, GraphNode, GraphSchemaModel


class NerEntitiesGraphConfig(GraphSchemaModel):
    """
    Configuration for the NER entities graph extractor.

    :ivar model: Named entity recognition model identifier.
    :vartype model: str
    :ivar min_entity_length: Minimum length for entity labels.
    :vartype min_entity_length: int
    :ivar include_item_node: Whether to emit an item node and mentions edges.
    :vartype include_item_node: bool
    """

    model_config = ConfigDict(extra="forbid")

    model: str = Field(min_length=1)
    min_entity_length: int = Field(default=3, ge=1)
    include_item_node: bool = Field(default=True)


class NerEntitiesGraphExtractor(GraphExtractor):
    """
    Graph extractor that emits named entity nodes and mentions edges.
    """

    extractor_id = "ner-entities"

    def validate_config(self, config: Dict[str, object]) -> BaseModel:
        """
        Validate configuration for NER entity extraction.

        :param config: Raw configuration mapping.
        :type config: dict[str, object]
        :return: Parsed configuration.
        :rtype: NerEntitiesGraphConfig
        """
        return NerEntitiesGraphConfig.model_validate(config)

    def extract_graph(
        self,
        *,
        corpus: Corpus,
        item: CatalogItem,
        extracted_text: str,
        config: BaseModel,
    ) -> GraphExtractionResult:
        """
        Extract graph nodes and edges for a single item.

        :param corpus: Corpus containing the item.
        :type corpus: Corpus
        :param item: Catalog item to extract from.
        :type item: CatalogItem
        :param extracted_text: Text to analyze.
        :type extracted_text: str
        :param config: Parsed configuration model.
        :type config: BaseModel
        :return: Graph extraction results.
        :rtype: GraphExtractionResult
        """
        _ = corpus
        parsed = config if isinstance(config, NerEntitiesGraphConfig) else None
        if parsed is None:
            parsed = NerEntitiesGraphConfig.model_validate(config)

        entities = _extract_entities(
            extracted_text=extracted_text,
            model_name=parsed.model,
            min_length=parsed.min_entity_length,
        )
        entity_counts = Counter(entity for entity, _ in entities)
        entity_types = {entity: label for entity, label in entities}

        nodes = _build_entity_nodes(entity_counts, entity_types)
        edges: List[GraphEdge] = []

        if parsed.include_item_node:
            item_node = GraphNode(
                node_id=f"item:{item.id}",
                node_type="item",
                label=item.title or item.relpath,
                properties={"item_id": item.id},
            )
            nodes.insert(0, item_node)
            edges.extend(_build_mentions_edges(item_node.node_id, entity_counts))

        return GraphExtractionResult(item_id=item.id, nodes=nodes, edges=edges)


def _extract_entities(
    *,
    extracted_text: str,
    model_name: str,
    min_length: int,
) -> List[Tuple[str, str]]:
    try:
        import spacy
    except ImportError as exc:
        raise ValueError(
            "NER graph extraction requires spaCy. Install it with pip install spacy."
        ) from exc
    nlp = spacy.load(model_name)
    doc = nlp(extracted_text)
    entities: List[Tuple[str, str]] = []
    for ent in getattr(doc, "ents", []):
        label = getattr(ent, "label_", "ENTITY")
        text = ent.text.strip()
        if len(text) < min_length:
            continue
        entities.append((text, label))
    return entities


def _canonicalize(label: str) -> str:
    lowered = label.lower()
    normalized = re.sub(r"[^a-z0-9]+", "_", lowered)
    normalized = re.sub(r"_+", "_", normalized).strip("_")
    return normalized or lowered


def _build_entity_nodes(
    entity_counts: Counter[str], entity_types: Dict[str, str]
) -> List[GraphNode]:
    nodes: List[GraphNode] = []
    for label in sorted(entity_counts.keys()):
        canonical = _canonicalize(label)
        nodes.append(
            GraphNode(
                node_id=f"entity:{canonical}",
                node_type="entity",
                label=label,
                properties={
                    "canonical": canonical,
                    "entity_type": entity_types.get(label, "ENTITY"),
                },
            )
        )
    return nodes


def _build_mentions_edges(item_node_id: str, entity_counts: Counter[str]) -> List[GraphEdge]:
    edges: List[GraphEdge] = []
    for label, count in sorted(entity_counts.items()):
        canonical = _canonicalize(label)
        entity_id = f"entity:{canonical}"
        edge_id = f"{item_node_id}|mentions|{entity_id}"
        edges.append(
            GraphEdge(
                edge_id=edge_id,
                src=item_node_id,
                dst=entity_id,
                edge_type="mentions",
                weight=float(count),
                properties={},
            )
        )
    return edges



"""
Data models for Reinforcement Memory.
"""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import Any, Dict, List, Optional


@dataclass
class TimestampedText:
    """
    A single text with a timestamp and optional metadata.

    :ivar id: Stable unique identifier for this text record.
    :vartype id: str
    :ivar group_id: Partitions texts into independent analysis groups.
    :vartype group_id: str
    :ivar timestamp: ISO 8601 timestamp (e.g. ``2024-01-15T10:00:00Z``).
    :vartype timestamp: str
    :ivar text: The text content to analyze.
    :vartype text: str
    :ivar metadata: Arbitrary key-value metadata preserved alongside the text.
    :vartype metadata: dict[str, Any]
    """

    id: str
    group_id: str
    timestamp: str
    text: str
    metadata: Dict[str, Any] = field(default_factory=dict)


@dataclass
class ExemplarRecord:
    """
    A representative example from a topic cluster.

    :ivar text: Truncated text of the exemplar.
    :vartype text: str
    :ivar text_id: ID of the source :class:`TimestampedText`.
    :vartype text_id: str
    :ivar metadata: Metadata from the source text.
    :vartype metadata: dict[str, Any]
    """

    text: str
    text_id: str
    metadata: Dict[str, Any] = field(default_factory=dict)


@dataclass
class TopicResult:
    """
    A discovered topic cluster with memory state.

    :ivar topic_id: Integer cluster identifier.
    :vartype topic_id: int
    :ivar label: Human-readable topic label (LLM-generated or keyword-based).
    :vartype label: str
    :ivar keywords: Top keywords characterizing the cluster.
    :vartype keywords: list[str]
    :ivar exemplars: Representative text samples closest to the centroid.
    :vartype exemplars: list[ExemplarRecord]
    :ivar member_count: Number of texts in this cluster.
    :vartype member_count: int
    :ivar memory_weight: Reinforcement weight in [0.0, 1.0].
    :vartype memory_weight: float
    :ivar memory_tier: ``hot``, ``warm``, or ``cold`` based on weight.
    :vartype memory_tier: str
    :ivar lifecycle_tier: ``new``, ``trending``, or ``established`` based on timestamps.
    :vartype lifecycle_tier: str
    :ivar is_new: True when cluster has only short-term members.
    :vartype is_new: bool
    :ivar is_trending: True when cluster has recent members but no long-term ones.
    :vartype is_trending: bool
    :ivar days_inactive: Days since the most recent member timestamp, or None.
    :vartype days_inactive: int | None
    :ivar root_cause: LLM-inferred root cause statement, or None.
    :vartype root_cause: str | None
    """

    topic_id: int
    label: str
    keywords: List[str]
    exemplars: List[ExemplarRecord]
    member_count: int
    memory_weight: float
    memory_tier: str
    lifecycle_tier: str
    is_new: bool
    is_trending: bool
    days_inactive: Optional[int]
    root_cause: Optional[str]


@dataclass
class AnalysisResult:
    """
    Result of a single :meth:`ReinforcementMemory.analyze` run.

    :ivar group_id: Group identifier this analysis covers.
    :vartype group_id: str
    :ivar topics: Discovered topic clusters, ordered by topic_id.
    :vartype topics: list[TopicResult]
    :ivar texts_analyzed: Number of texts included in this analysis.
    :vartype texts_analyzed: int
    :ivar cluster_version: Timestamp string identifying this clustering run.
    :vartype cluster_version: str
    :ivar run_id: Unique identifier for this analysis run.
    :vartype run_id: str
    """

    group_id: str
    topics: List[TopicResult]
    texts_analyzed: int
    cluster_version: str
    run_id: str


@dataclass
class VectorRecord:
    """
    A single vector with metadata for storage in a :class:`VectorStore`.

    :ivar key: Unique key identifying this vector.
    :vartype key: str
    :ivar embedding: Float32 embedding vector.
    :vartype embedding: list[float]
    :ivar metadata: Arbitrary metadata stored alongside the vector.
    :vartype metadata: dict[str, Any]
    """

    key: str
    embedding: List[float]
    metadata: Dict[str, Any] = field(default_factory=dict)


@dataclass
class QueryResult:
    """
    A single result from a vector similarity query.

    :ivar key: Key of the matched vector.
    :vartype key: str
    :ivar distance: Cosine distance (lower = more similar).
    :vartype distance: float
    :ivar similarity: Cosine similarity = 1 - distance.
    :vartype similarity: float
    :ivar metadata: Metadata stored with the matched vector.
    :vartype metadata: dict[str, Any]
    """

    key: str
    distance: float
    similarity: float
    metadata: Dict[str, Any] = field(default_factory=dict)



from __future__ import annotations

import os
import sys
import types
from dataclasses import dataclass
from typing import Any, Dict, Optional

from behave import given, then, when

from biblicus.user_config import load_user_config, resolve_deepgram_api_key


@dataclass
class _FakeDeepgramTranscriptionBehavior:
    transcript: Optional[str]
    response_type: str = "normal"


def _ensure_fake_deepgram_transcription_behaviors(
    context,
) -> Dict[str, _FakeDeepgramTranscriptionBehavior]:
    behaviors = getattr(context, "fake_deepgram_transcriptions", None)
    if behaviors is None:
        behaviors = {}
        context.fake_deepgram_transcriptions = behaviors
    return behaviors


def _install_fake_deepgram_module(context) -> None:
    deepgram_module = sys.modules.get("deepgram")
    if deepgram_module is not None and hasattr(deepgram_module, "DeepgramClient"):
        return

    original_modules: Dict[str, object] = {}
    module_names = [
        "deepgram",
    ]
    for name in module_names:
        if name in sys.modules:
            original_modules[name] = sys.modules[name]

    class _Alternative:
        def __init__(self, transcript: str) -> None:
            self.transcript = transcript
            words = [
                {"word": word, "channel": 0, "speaker": 0}
                for word in transcript.split()
                if word
            ]
            self.words = words
            self.utterances = [
                {
                    "transcript": transcript,
                    "channel": 0,
                    "speaker": 0,
                }
            ]

    class _Channel:
        def __init__(self, alternatives: list) -> None:
            self.alternatives = alternatives

    class _Results:
        def __init__(self, channels: list) -> None:
            self.channels = channels
            alternatives = channels[0].alternatives if channels else []
            utterances = []
            words = []
            if alternatives:
                utterances = alternatives[0].utterances
                words = alternatives[0].words
            self.utterances = utterances
            self.words = words

    class _TranscriptionResponseNormal:
        def __init__(self, transcript: str) -> None:
            self.results = _Results([_Channel([_Alternative(transcript)])])

        def to_dict(self) -> Dict[str, Any]:
            alternatives = self.results.channels[0].alternatives if self.results.channels else []
            payload = {"results": {"channels": []}}
            if alternatives:
                alternative = alternatives[0]
                payload["results"]["channels"].append(
                    {
                        "alternatives": [
                            {
                                "transcript": alternative.transcript,
                                "words": alternative.words,
                                "utterances": alternative.utterances,
                            }
                        ]
                    }
                )
                payload["results"]["utterances"] = alternative.utterances
                payload["results"]["words"] = alternative.words
            return payload

    class _TranscriptionResponseNoResults:
        def __init__(self) -> None:
            self.results = None

        def to_dict(self) -> Dict[str, Any]:
            return {"results": None}

    class _TranscriptionResponseEmptyChannels:
        def __init__(self) -> None:
            self.results = _Results([])

        def to_dict(self) -> Dict[str, Any]:
            return {"results": {"channels": []}}

    class _TranscriptionResponseEmptyAlternatives:
        def __init__(self) -> None:
            self.results = _Results([_Channel([])])

        def to_dict(self) -> Dict[str, Any]:
            return {"results": {"channels": [{"alternatives": []}]}}

    class _TranscribeApi:
        def transcribe_file(self, *args: Any, **kwargs: Any) -> object:
            options = {key: value for key, value in kwargs.items() if key != "request"}
            deepgram_module.last_transcription_model = options.get("model")
            deepgram_module.last_transcription_options = dict(options)
            behaviors_map = _ensure_fake_deepgram_transcription_behaviors(context)
            for _filename, behavior in behaviors_map.items():
                if behavior.response_type == "empty_results":
                    return _TranscriptionResponseNoResults()
                if behavior.response_type == "empty_channels":
                    return _TranscriptionResponseEmptyChannels()
                if behavior.response_type == "empty_alternatives":
                    return _TranscriptionResponseEmptyAlternatives()
                if behavior.transcript is not None:
                    return _TranscriptionResponseNormal(behavior.transcript)
            return _TranscriptionResponseNormal("")

    class _ListenV1Media:
        def transcribe_file(self, *args: Any, **kwargs: Any) -> object:
            return _TranscribeApi().transcribe_file(*args, **kwargs)

    class _ListenV1:
        def __init__(self) -> None:
            self.media = _ListenV1Media()

    class _Listen:
        def __init__(self) -> None:
            self.v1 = _ListenV1()

    class DeepgramClient:
        def __init__(self, api_key: str = "", **kwargs: Any) -> None:
            deepgram_module.last_api_key = api_key
            self.listen = _Listen()

    deepgram_module = types.ModuleType("deepgram")
    deepgram_module.DeepgramClient = DeepgramClient
    deepgram_module.last_api_key = None
    deepgram_module.last_transcription_model = None
    deepgram_module.last_transcription_options = {}

    sys.modules["deepgram"] = deepgram_module

    context._fake_deepgram_original_modules = original_modules


def _install_deepgram_unavailable_module(context) -> None:
    deepgram_module = sys.modules.get("deepgram")
    if deepgram_module is not None and not hasattr(deepgram_module, "DeepgramClient"):
        return

    original_modules: Dict[str, object] = {}
    module_names = [
        "deepgram",
    ]
    for name in module_names:
        if name in sys.modules:
            original_modules[name] = sys.modules[name]

    deepgram_module = types.ModuleType("deepgram")
    sys.modules["deepgram"] = deepgram_module

    context._fake_deepgram_unavailable_original_modules = original_modules


@given("a fake Deepgram library is available")
def step_fake_deepgram_available(context) -> None:
    _install_fake_deepgram_module(context)


@given(
    'a fake Deepgram library is available that returns transcript "{transcript}" for filename "{filename}"'
)
def step_fake_deepgram_returns_transcript(context, transcript: str, filename: str) -> None:
    _install_fake_deepgram_module(context)
    behaviors = _ensure_fake_deepgram_transcription_behaviors(context)
    behaviors[filename] = _FakeDeepgramTranscriptionBehavior(transcript=transcript)


@given('a fake Deepgram library is available that returns empty results for filename "{filename}"')
def step_fake_deepgram_returns_empty_results(context, filename: str) -> None:
    _install_fake_deepgram_module(context)
    behaviors = _ensure_fake_deepgram_transcription_behaviors(context)
    behaviors[filename] = _FakeDeepgramTranscriptionBehavior(
        transcript=None, response_type="empty_results"
    )


@given('a fake Deepgram library is available that returns empty channels for filename "{filename}"')
def step_fake_deepgram_returns_empty_channels(context, filename: str) -> None:
    _install_fake_deepgram_module(context)
    behaviors = _ensure_fake_deepgram_transcription_behaviors(context)
    behaviors[filename] = _FakeDeepgramTranscriptionBehavior(
        transcript=None, response_type="empty_channels"
    )


@given(
    'a fake Deepgram library is available that returns empty alternatives for filename "{filename}"'
)
def step_fake_deepgram_returns_empty_alternatives(context, filename: str) -> None:
    _install_fake_deepgram_module(context)
    behaviors = _ensure_fake_deepgram_transcription_behaviors(context)
    behaviors[filename] = _FakeDeepgramTranscriptionBehavior(
        transcript=None, response_type="empty_alternatives"
    )


@given("a Deepgram API key is configured for this scenario")
@when("a Deepgram API key is configured for this scenario")
def step_deepgram_api_key_configured(context) -> None:
    api_key = resolve_deepgram_api_key()
    if not api_key and getattr(context, "repo_root", None) is not None:
        repo_root = context.repo_root
        candidate = repo_root / ".biblicus" / "config.yml"
        if candidate.is_file():
            loaded = load_user_config(paths=[candidate])
            if loaded.deepgram is not None:
                api_key = loaded.deepgram.api_key
    if not api_key:
        scenario = getattr(context, "scenario", None)
        tags = set(getattr(scenario, "tags", [])) if scenario else set()
        feature = getattr(context, "feature", None)
        if feature is not None:
            tags.update(getattr(feature, "tags", []))
        if "deepgram" in tags or "integration" in tags:
            if scenario is not None:
                scenario.skip("DEEPGRAM_API_KEY is required for Deepgram integration scenarios.")
            return
        api_key = "test-deepgram-key"
    extra_env = getattr(context, "extra_env", None)
    if extra_env is None:
        extra_env = {}
        context.extra_env = extra_env
    extra_env["DEEPGRAM_API_KEY"] = api_key


@given("the Deepgram dependency is unavailable")
def step_deepgram_dependency_unavailable(context) -> None:
    _install_deepgram_unavailable_module(context)


@then('the Deepgram transcription request used model "{model}"')
def step_deepgram_transcription_used_model(context, model: str) -> None:
    _ = context
    deepgram_module = sys.modules.get("deepgram")
    assert deepgram_module is not None
    assert getattr(deepgram_module, "last_transcription_model", None) == model


@then("the Deepgram transcription request used smart format true")
def step_deepgram_transcription_used_smart_format_true(context) -> None:
    _ = context
    deepgram_module = sys.modules.get("deepgram")
    assert deepgram_module is not None
    options: Dict[str, Any] = getattr(deepgram_module, "last_transcription_options", {})
    assert options.get("smart_format") is True


@then("the Deepgram transcription request used punctuate true")
def step_deepgram_transcription_used_punctuate_true(context) -> None:
    _ = context
    deepgram_module = sys.modules.get("deepgram")
    assert deepgram_module is not None
    options: Dict[str, Any] = getattr(deepgram_module, "last_transcription_options", {})
    assert options.get("punctuate") is True



"""
Dependency relation graph extractor for Biblicus.
"""

from __future__ import annotations

import re
from collections import Counter
from typing import Dict, Iterable, List, Tuple

from pydantic import BaseModel, ConfigDict, Field

from ...corpus import Corpus
from ...models import CatalogItem
from ..base import GraphExtractor
from ..models import GraphEdge, GraphExtractionResult, GraphNode, GraphSchemaModel


class DependencyRelationsGraphConfig(GraphSchemaModel):
    """
    Configuration for the dependency relations graph extractor.

    :ivar model: Dependency parsing model identifier.
    :vartype model: str
    :ivar min_entity_length: Minimum length for entity labels.
    :vartype min_entity_length: int
    :ivar include_item_node: Whether to emit an item node and mentions edges.
    :vartype include_item_node: bool
    """

    model_config = ConfigDict(extra="forbid")

    model: str = Field(min_length=1)
    min_entity_length: int = Field(default=3, ge=1)
    include_item_node: bool = Field(default=True)


class DependencyRelationsGraphExtractor(GraphExtractor):
    """
    Graph extractor that emits dependency-based relation edges.
    """

    extractor_id = "dependency-relations"

    def validate_config(self, config: Dict[str, object]) -> BaseModel:
        """
        Validate configuration for dependency relations extraction.

        :param config: Raw configuration mapping.
        :type config: dict[str, object]
        :return: Parsed configuration.
        :rtype: DependencyRelationsGraphConfig
        """
        return DependencyRelationsGraphConfig.model_validate(config)

    def extract_graph(
        self,
        *,
        corpus: Corpus,
        item: CatalogItem,
        extracted_text: str,
        config: BaseModel,
    ) -> GraphExtractionResult:
        """
        Extract graph nodes and edges for a single item.

        :param corpus: Corpus containing the item.
        :type corpus: Corpus
        :param item: Catalog item to extract from.
        :type item: CatalogItem
        :param extracted_text: Text to analyze.
        :type extracted_text: str
        :param config: Parsed configuration model.
        :type config: BaseModel
        :return: Graph extraction results.
        :rtype: GraphExtractionResult
        """
        _ = corpus
        parsed = config if isinstance(config, DependencyRelationsGraphConfig) else None
        if parsed is None:
            parsed = DependencyRelationsGraphConfig.model_validate(config)

        entities = _extract_entities(
            extracted_text=extracted_text,
            model_name=parsed.model,
            min_length=parsed.min_entity_length,
        )
        relations = _extract_relations(
            extracted_text=extracted_text,
            model_name=parsed.model,
            min_length=parsed.min_entity_length,
        )
        entity_counts = Counter(entity for entity, _ in entities)
        entity_types = {entity: label for entity, label in entities}

        nodes = _build_entity_nodes(entity_counts, entity_types)
        edges: List[GraphEdge] = []

        if parsed.include_item_node:
            item_node = GraphNode(
                node_id=f"item:{item.id}",
                node_type="item",
                label=item.title or item.relpath,
                properties={"item_id": item.id},
            )
            nodes.insert(0, item_node)
            edges.extend(_build_mentions_edges(item_node.node_id, entity_counts))

        edges.extend(_build_relation_edges(relations))

        return GraphExtractionResult(item_id=item.id, nodes=nodes, edges=edges)


def _extract_entities(
    *,
    extracted_text: str,
    model_name: str,
    min_length: int,
) -> List[Tuple[str, str]]:
    doc = _load_doc(extracted_text, model_name)
    entities: List[Tuple[str, str]] = []
    for ent in getattr(doc, "ents", []):
        label = getattr(ent, "label_", "ENTITY")
        text = ent.text.strip()
        if len(text) < min_length:
            continue
        entities.append((text, label))
    return entities


def _extract_relations(
    *,
    extracted_text: str,
    model_name: str,
    min_length: int,
) -> List[Tuple[str, str, str]]:
    doc = _load_doc(extracted_text, model_name)
    relations: List[Tuple[str, str, str]] = []
    subject_deps = {"nsubj", "nsubjpass"}
    object_deps = {"dobj", "obj", "pobj"}

    for token in doc:
        if token.dep_ not in subject_deps:
            continue
        verb = token.head
        objects = [
            child for child in getattr(verb, "children", []) if child.dep_ in object_deps
        ]
        subj_text = token.text.strip()
        if len(subj_text) < min_length:
            continue
        for obj in objects:
            obj_text = obj.text.strip()
            if len(obj_text) < min_length:
                continue
            predicate = getattr(verb, "lemma_", verb.text)
            relations.append((subj_text, predicate, obj_text))
    return relations


def _load_doc(text: str, model_name: str):
    try:
        import spacy
    except ImportError as exc:
        raise ValueError(
            "Dependency graph extraction requires spaCy. Install it with pip install spacy."
        ) from exc
    nlp = spacy.load(model_name)
    return nlp(text)


def _canonicalize(label: str) -> str:
    lowered = label.lower()
    normalized = re.sub(r"[^a-z0-9]+", "_", lowered)
    normalized = re.sub(r"_+", "_", normalized).strip("_")
    return normalized or lowered


def _build_entity_nodes(
    entity_counts: Counter[str], entity_types: Dict[str, str]
) -> List[GraphNode]:
    nodes: List[GraphNode] = []
    for label in sorted(entity_counts.keys()):
        canonical = _canonicalize(label)
        nodes.append(
            GraphNode(
                node_id=f"entity:{canonical}",
                node_type="entity",
                label=label,
                properties={
                    "canonical": canonical,
                    "entity_type": entity_types.get(label, "ENTITY"),
                },
            )
        )
    return nodes


def _build_mentions_edges(item_node_id: str, entity_counts: Counter[str]) -> List[GraphEdge]:
    edges: List[GraphEdge] = []
    for label, count in sorted(entity_counts.items()):
        canonical = _canonicalize(label)
        entity_id = f"entity:{canonical}"
        edge_id = f"{item_node_id}|mentions|{entity_id}"
        edges.append(
            GraphEdge(
                edge_id=edge_id,
                src=item_node_id,
                dst=entity_id,
                edge_type="mentions",
                weight=float(count),
                properties={},
            )
        )
    return edges


def _build_relation_edges(relations: Iterable[Tuple[str, str, str]]) -> List[GraphEdge]:
    edges: List[GraphEdge] = []
    for subj, predicate, obj in relations:
        src = f"entity:{_canonicalize(subj)}"
        dst = f"entity:{_canonicalize(obj)}"
        edge_id = f"{src}|related_to|{dst}|{_canonicalize(predicate)}"
        edges.append(
            GraphEdge(
                edge_id=edge_id,
                src=src,
                dst=dst,
                edge_type="related_to",
                weight=1.0,
                properties={"predicate": predicate},
            )
        )
    return edges



"""
Memory weights: reinforcement and decay for topic clusters.

Clusters that receive new documents are reinforced (weight -> 1.0).
Clusters with no new documents decay (weight -> 0.0).
Tiers: hot >= 0.7, warm >= 0.3, cold < 0.3.
"""

from typing import Any, Dict, List, Optional, Tuple

DEFAULT_HOT_THRESHOLD = 0.7
DEFAULT_WARM_THRESHOLD = 0.3
DEFAULT_PRUNE_THRESHOLD = 0.1


def initial_weight() -> float:
    """Return neutral starting weight for new clusters."""
    return 0.5


def reinforce(weight: float, new_docs: int, rate: float = 0.1) -> float:
    """
    Increase weight toward 1.0 based on new document count.

    :param weight: Current weight in [0.0, 1.0].
    :param new_docs: Number of new documents matched to this cluster.
    :param rate: Maximum weight increase per reinforcement step.
    :return: Updated weight, clamped to [0.0, 1.0].
    """
    delta = rate * min(new_docs / 10.0, 1.0)
    return min(1.0, weight + delta)


def decay(weight: float, days: int, rate: float = 0.05) -> float:
    """
    Decrease weight toward 0.0 based on days since last reinforcement.

    :param weight: Current weight in [0.0, 1.0].
    :param days: Days elapsed since this cluster last received new documents.
    :param rate: Maximum weight decrease per decay step.
    :return: Updated weight, clamped to [0.0, 1.0].
    """
    delta = rate * min(days / 7.0, 1.0)
    return max(0.0, weight - delta)


def tier_from_weight(
    weight: float,
    hot_threshold: float = DEFAULT_HOT_THRESHOLD,
    warm_threshold: float = DEFAULT_WARM_THRESHOLD,
) -> str:
    """
    Map a weight value to a memory tier label.

    :param weight: Weight in [0.0, 1.0].
    :param hot_threshold: Minimum weight for ``hot`` tier.
    :param warm_threshold: Minimum weight for ``warm`` tier.
    :return: ``"hot"``, ``"warm"``, or ``"cold"``.
    """
    if weight >= hot_threshold:
        return "hot"
    if weight >= warm_threshold:
        return "warm"
    return "cold"


def should_prune(weight: float, threshold: float = DEFAULT_PRUNE_THRESHOLD) -> bool:
    """
    Return True if weight is below the prune threshold.

    :param weight: Weight in [0.0, 1.0].
    :param threshold: Prune threshold.
    :return: True if the cluster should be pruned.
    """
    return weight < threshold


def update_memory_weights(
    existing_clusters: List[Dict[str, Any]],
    active_cluster_ids: List[int],
    days_inactive: Optional[Dict[int, int]] = None,
    reinforce_rate: float = 0.1,
    decay_rate: float = 0.05,
    prune_threshold: float = DEFAULT_PRUNE_THRESHOLD,
    prune: bool = True,
) -> Tuple[List[Dict[str, Any]], List[int]]:
    """
    Compute updated weights and tiers for a list of cluster records.

    Clusters present in ``active_cluster_ids`` are reinforced; all others
    are decayed.  Clusters whose weight falls below ``prune_threshold`` are
    removed from the result and their IDs are returned in the pruned list.

    :param existing_clusters: List of cluster dicts, each with at least a
        ``cluster_id`` and ``memory_weight`` field.
    :param active_cluster_ids: IDs of clusters that received new documents.
    :param days_inactive: Mapping of cluster_id -> days since last activity,
        used to scale decay.  Defaults to 7 days for unknown clusters.
    :param reinforce_rate: Rate for :func:`reinforce`.
    :param decay_rate: Rate for :func:`decay`.
    :param prune_threshold: Clusters below this weight are pruned.
    :param prune: If False, pruning is skipped.
    :return: ``(updated_clusters, pruned_cluster_ids)`` tuple.
    """
    active_set = set(active_cluster_ids)
    days_inactive = days_inactive or {}
    updated: List[Dict[str, Any]] = []
    pruned: List[int] = []

    for c in existing_clusters:
        cid = c.get("cluster_id")
        if cid is None:
            continue
        cid_int = int(cid) if isinstance(cid, str) and cid.lstrip("-").isdigit() else cid
        weight = c.get("memory_weight", initial_weight())
        if isinstance(weight, (int, float)):
            weight = float(weight)
        else:
            weight = initial_weight()

        if cid_int in active_set:
            new_docs = c.get("new_docs_this_run", 10)
            weight = reinforce(weight, new_docs, reinforce_rate)
        else:
            days = days_inactive.get(cid_int, 7)
            weight = decay(weight, days, decay_rate)

        weight = max(0.0, min(1.0, weight))
        tier = tier_from_weight(weight)

        if prune and should_prune(weight, prune_threshold):
            pruned.append(cid_int)
            continue

        updated.append(
            {
                **c,
                "memory_weight": weight,
                "memory_tier": tier,
            }
        )

    return updated, pruned



"""
SmolDocling VLM-backed document text extraction plugin.

This extractor uses the SmolDocling-256M vision-language model for document understanding.
It supports PDF, Office documents (DOCX, XLSX, PPTX), HTML, and image formats.
"""

from __future__ import annotations

from typing import Any, Dict, List, Optional

from pydantic import BaseModel, ConfigDict, Field

from ..corpus import Corpus
from ..errors import ExtractionSnapshotFatalError
from ..models import CatalogItem, ExtractedText, ExtractionStageOutput
from .base import TextExtractor

DOCLING_SUPPORTED_MEDIA_TYPES = frozenset(
    [
        "application/pdf",
        "application/vnd.openxmlformats-officedocument.wordprocessingml.document",
        "application/vnd.openxmlformats-officedocument.spreadsheetml.sheet",
        "application/vnd.openxmlformats-officedocument.presentationml.presentation",
        "text/html",
        "application/xhtml+xml",
        "image/png",
        "image/jpeg",
        "image/gif",
        "image/webp",
        "image/tiff",
        "image/bmp",
    ]
)


class DoclingSmolExtractorConfig(BaseModel):
    """
    Configuration for the SmolDocling VLM extractor.

    :ivar output_format: Output format for extracted content (markdown, text, or html).
    :vartype output_format: str
    :ivar retriever: Inference retriever (mlx or transformers).
    :vartype retriever: str
    """

    model_config = ConfigDict(extra="forbid", populate_by_name=True)

    output_format: str = Field(default="markdown", pattern="^(markdown|text|html)$")
    retriever: str = Field(default="mlx", pattern="^(mlx|transformers)$", alias="backend")


class DoclingSmolExtractor(TextExtractor):
    """
    Extractor plugin backed by the SmolDocling-256M vision-language model.

    This extractor converts documents into text using Docling with the SmolDocling VLM.
    It skips text items (text/plain, text/markdown) to let pass-through handle those.

    :ivar extractor_id: Extractor identifier.
    :vartype extractor_id: str
    """

    extractor_id = "docling-smol"

    def validate_config(self, config: Dict[str, Any]) -> BaseModel:
        """
        Validate extractor configuration and ensure the dependency is installed.

        :param config: Configuration mapping.
        :type config: dict[str, Any]
        :return: Parsed config.
        :rtype: DoclingSmolExtractorConfig
        :raises ExtractionSnapshotFatalError: If the optional dependency is not installed.
        """
        parsed = DoclingSmolExtractorConfig.model_validate(config)

        try:
            from docling.document_converter import DocumentConverter  # noqa: F401
            # Updated import path for docling 2.x
            try:
                from docling.datamodel.pipeline_options import (  # noqa: F401
                    PdfPipelineOptions,
                )
            except ImportError:
                # Fallback for older docling versions
                from docling.pipeline_options import (  # noqa: F401
                    VlmPipelineOptions,
                    vlm_model_specs,
                )
        except ImportError as import_error:
            raise ExtractionSnapshotFatalError(
                "DoclingSmol extractor requires an optional dependency. "
                'Install it with pip install "biblicus[docling]".'
            ) from import_error

        if parsed.retriever == "mlx":
            try:
                try:
                    # Try new API first
                    from docling.datamodel.pipeline_options import PdfPipelineOptions
                    _ = PdfPipelineOptions
                except ImportError:
                    # Fallback to old API
                    from docling.pipeline_options import vlm_model_specs
                    _ = vlm_model_specs.SMOLDOCLING_MLX
            except (ImportError, AttributeError) as exc:
                raise ExtractionSnapshotFatalError(
                    "DoclingSmol extractor with MLX retriever requires MLX support. "
                    'Install it with pip install "biblicus[docling-mlx]".'
                ) from exc

        return parsed

    def extract_text(
        self,
        *,
        corpus: Corpus,
        item: CatalogItem,
        config: BaseModel,
        previous_extractions: List[ExtractionStageOutput],
    ) -> Optional[ExtractedText]:
        """
        Extract text for a document item using SmolDocling.

        :param corpus: Corpus containing the item bytes.
        :type corpus: Corpus
        :param item: Catalog item being processed.
        :type item: CatalogItem
        :param config: Parsed configuration model.
        :type config: DoclingSmolExtractorConfig
        :param previous_extractions: Prior stage outputs for this item within the pipeline.
        :type previous_extractions: list[biblicus.models.ExtractionStageOutput]
        :return: Extracted text payload, or None when the item is not supported.
        :rtype: ExtractedText or None
        """
        _ = previous_extractions

        if not self._is_supported_media_type(item.media_type):
            return None

        parsed_config = (
            config
            if isinstance(config, DoclingSmolExtractorConfig)
            else DoclingSmolExtractorConfig.model_validate(config)
        )

        source_path = corpus.root / item.relpath
        text = self._convert_document(source_path, parsed_config)
        return ExtractedText(text=text.strip(), producer_extractor_id=self.extractor_id)

    def _is_supported_media_type(self, media_type: str) -> bool:
        """
        Check if a media type is supported by this extractor.

        :param media_type: Media type string.
        :type media_type: str
        :return: True if supported, False otherwise.
        :rtype: bool
        """
        if media_type in DOCLING_SUPPORTED_MEDIA_TYPES:
            return True
        if media_type.startswith("image/"):
            return True
        return False

    def _convert_document(self, source_path, config: DoclingSmolExtractorConfig) -> str:
        """
        Convert a document using Docling with the SmolDocling VLM.

        :param source_path: Path to the source document.
        :type source_path: pathlib.Path
        :param config: Parsed configuration.
        :type config: DoclingSmolExtractorConfig
        :return: Extracted text content.
        :rtype: str
        """
        from docling.document_converter import DocumentConverter

        # Current Docling API is simplified - just create converter and convert
        # The VLM model configuration (MLX vs Transformers) is handled via environment
        # or model selection, not via explicit pipeline options
        converter = DocumentConverter()
        result = converter.convert(str(source_path))

        if config.output_format == "html":
            return result.document.export_to_html()
        elif config.output_format == "text":
            return result.document.export_to_text()
        else:
            return result.document.export_to_markdown()



"""
Inference backend abstraction for machine learning powered components.

This module provides reusable configuration and credential resolution patterns for components
that can execute locally or via API providers.
"""

from __future__ import annotations

import os
from enum import Enum
from typing import Optional

from pydantic import BaseModel, ConfigDict, Field, model_validator


class InferenceBackendMode(str, Enum):
    """Execution mode for inference backends."""

    LOCAL = "local"
    API = "api"


class ApiProvider(str, Enum):
    """Supported application programming interface providers for inference."""

    HUGGINGFACE = "huggingface"
    OPENAI = "openai"


class InferenceBackendConfig(BaseModel):
    """
    Composable configuration for inference backends.

    This config can be embedded in extractor or transformer configurations to provide
    a uniform interface for local versus application programming interface execution.

    :ivar mode: Execution mode, local or application programming interface.
    :vartype mode: InferenceBackendMode
    :ivar api_provider: Application programming interface provider when mode is application programming interface.
    :vartype api_provider: ApiProvider or None
    :ivar api_key: Optional per-config application programming interface key override.
    :vartype api_key: str or None
    :ivar model_id: Optional model identifier for application programming interface requests.
    :vartype model_id: str or None
    """

    model_config = ConfigDict(extra="forbid")

    mode: InferenceBackendMode = Field(default=InferenceBackendMode.LOCAL)
    api_provider: Optional[ApiProvider] = Field(default=None)
    api_key: Optional[str] = Field(default=None)
    model_id: Optional[str] = Field(default=None)

    @model_validator(mode="after")
    def _validate_api_provider_required(self) -> "InferenceBackendConfig":
        if self.mode == InferenceBackendMode.API and self.api_provider is None:
            raise ValueError("api_provider is required when mode is 'api'")
        return self


def resolve_api_key(
    provider: ApiProvider,
    *,
    config_override: Optional[str] = None,
) -> Optional[str]:
    """
    Resolve an application programming interface key with precedence rules.

    Precedence order (highest to lowest):
    1. Explicit config override parameter
    2. Environment variable for the provider
    3. User configuration file

    :param provider: Application programming interface provider to resolve key for.
    :type provider: ApiProvider
    :param config_override: Optional explicit key from configuration.
    :type config_override: str or None
    :return: Resolved application programming interface key or None if unavailable.
    :rtype: str or None
    """
    if config_override is not None:
        return config_override

    from .user_config import load_user_config

    if provider == ApiProvider.HUGGINGFACE:
        env_key = os.environ.get("HUGGINGFACE_API_KEY")
        if env_key:
            return env_key
        user_config = load_user_config()
        if user_config.huggingface is not None:
            return user_config.huggingface.api_key
        return None
    elif provider == ApiProvider.OPENAI:
        env_key = os.environ.get("OPENAI_API_KEY")
        if env_key:
            return env_key
        user_config = load_user_config()
        if user_config.openai is not None:
            return user_config.openai.api_key
        return None
    else:
        return None



"""
Prepare scanned research paper samples for layout-aware OCR testing.

This script helps convert PDFs with selectable text into image-based PDFs
that simulate scanned documents, perfect for testing OCR pipelines.
"""

from __future__ import annotations

import argparse
import json
from pathlib import Path
from typing import Dict, List

from biblicus.corpus import Corpus


def pdf_to_images(pdf_path: Path, output_dir: Path, dpi: int = 300) -> List[Path]:
    """
    Convert PDF pages to PNG images.

    :param pdf_path: Path to input PDF.
    :type pdf_path: Path
    :param output_dir: Directory for output images.
    :type output_dir: Path
    :param dpi: Resolution in dots per inch.
    :type dpi: int
    :return: List of created image paths.
    :rtype: list[Path]
    :raises ImportError: If pdf2image is not installed.
    """
    try:
        from pdf2image import convert_from_path
    except ImportError as e:
        raise ImportError(
            "pdf2image is required. Install with: pip install pdf2image\n"
            "Also requires poppler: brew install poppler (macOS) or "
            "apt-get install poppler-utils (Ubuntu)"
        ) from e

    output_dir.mkdir(parents=True, exist_ok=True)

    print(f"Converting PDF to images at {dpi} DPI...")
    images = convert_from_path(pdf_path, dpi=dpi)

    image_paths = []
    for i, image in enumerate(images, start=1):
        output_path = output_dir / f"{pdf_path.stem}_page_{i:03d}.png"
        image.save(output_path, "PNG")
        image_paths.append(output_path)
        print(f"  ✓ Page {i} → {output_path.name}")

    return image_paths


def images_to_pdf(image_paths: List[Path], output_pdf: Path) -> Path:
    """
    Combine PNG images into a single PDF (simulating scanned document).

    :param image_paths: List of image file paths.
    :type image_paths: list[Path]
    :param output_pdf: Output PDF path.
    :type output_pdf: Path
    :return: Path to created PDF.
    :rtype: Path
    :raises ImportError: If PIL is not installed.
    """
    try:
        from PIL import Image
    except ImportError as e:
        raise ImportError("Pillow is required. Install with: pip install Pillow") from e

    output_pdf.parent.mkdir(parents=True, exist_ok=True)

    print(f"Creating image-based PDF: {output_pdf.name}")

    # Open all images
    images = [Image.open(img_path).convert("RGB") for img_path in image_paths]

    # Save as PDF
    if len(images) == 1:
        images[0].save(output_pdf, "PDF", resolution=100.0)
    else:
        images[0].save(output_pdf, "PDF", resolution=100.0, save_all=True, append_images=images[1:])

    print(f"  ✓ Created {output_pdf}")
    return output_pdf


def prepare_scanned_sample(
    input_pdf: Path,
    output_dir: Path,
    corpus_path: Path | None = None,
    tags: List[str] | None = None,
    dpi: int = 300,
    keep_images: bool = False,
) -> Dict[str, object]:
    """
    Convert a PDF into a scanned-like image PDF and optionally ingest to corpus.

    :param input_pdf: Source PDF file.
    :type input_pdf: Path
    :param output_dir: Directory for outputs.
    :type output_dir: Path
    :param corpus_path: Optional corpus to ingest into.
    :type corpus_path: Path | None
    :param tags: Tags to apply when ingesting.
    :type tags: list[str] | None
    :param dpi: Resolution for conversion.
    :type dpi: int
    :param keep_images: Whether to keep intermediate PNG files.
    :type keep_images: bool
    :return: Processing statistics.
    :rtype: dict[str, object]
    """
    print("=" * 70)
    print(f"PREPARING SCANNED SAMPLE: {input_pdf.name}")
    print("=" * 70)
    print()

    # Create temp directory for images
    images_dir = output_dir / "temp_images" / input_pdf.stem

    # Step 1: PDF → PNG images
    image_paths = pdf_to_images(input_pdf, images_dir, dpi=dpi)
    print(f"\n✓ Converted {len(image_paths)} pages to images")

    # Step 2: PNG images → image-based PDF
    output_pdf = output_dir / f"{input_pdf.stem}_scanned.pdf"
    images_to_pdf(image_paths, output_pdf)
    print(f"✓ Created image-based PDF: {output_pdf}")

    # Step 3: Optionally ingest to corpus
    item_id = None
    if corpus_path:
        print(f"\nIngesting to corpus: {corpus_path}")
        corpus = Corpus.open(corpus_path) if corpus_path.exists() else Corpus.init(corpus_path)
        result = corpus.ingest_source(output_pdf, tags=tags or ["scanned", "multi-column"])
        item_id = result.item_id
        print(f"✓ Ingested as: {item_id}")

    # Step 4: Cleanup intermediate files
    if not keep_images:
        import shutil

        shutil.rmtree(images_dir)
        print("✓ Cleaned up temporary images")

    print()
    return {
        "input_pdf": str(input_pdf),
        "output_pdf": str(output_pdf),
        "pages": len(image_paths),
        "item_id": item_id,
        "images_kept": keep_images,
    }


def build_parser() -> argparse.ArgumentParser:
    """Build argument parser."""
    parser = argparse.ArgumentParser(
        description="Convert PDFs to scanned-like images for OCR testing"
    )
    parser.add_argument("input_pdf", type=Path, help="Input PDF file to convert")
    parser.add_argument(
        "--output-dir",
        type=Path,
        default=Path("scanned_samples"),
        help="Output directory (default: scanned_samples/)",
    )
    parser.add_argument("--corpus", type=Path, help="Corpus to ingest the result into")
    parser.add_argument("--tag", action="append", dest="tags", help="Tag to apply (repeatable)")
    parser.add_argument("--dpi", type=int, default=300, help="DPI resolution (default: 300)")
    parser.add_argument("--keep-images", action="store_true", help="Keep intermediate PNG files")
    return parser


def main() -> int:
    """Entry point."""
    parser = build_parser()
    args = parser.parse_args()

    if not args.input_pdf.exists():
        print(f"ERROR: Input PDF not found: {args.input_pdf}")
        return 1

    result = prepare_scanned_sample(
        input_pdf=args.input_pdf,
        output_dir=args.output_dir,
        corpus_path=args.corpus,
        tags=args.tags,
        dpi=args.dpi,
        keep_images=args.keep_images,
    )

    print("=" * 70)
    print("RESULT")
    print("=" * 70)
    print(json.dumps(result, indent=2))
    return 0


if __name__ == "__main__":
    raise SystemExit(main())



import pytest
from types import SimpleNamespace

from biblicus.analysis import markov
from biblicus.analysis.models import (
    MarkovAnalysisConfiguration,
    MarkovAnalysisObservationsConfig,
    MarkovAnalysisObservationsEncoder,
    MarkovAnalysisModelConfig,
    MarkovAnalysisModelFamily,
    MarkovAnalysisSegment,
    MarkovAnalysisTopicModelingConfig,
    MarkovAnalysisSegmentationConfig,
    MarkovAnalysisLlmObservationsConfig,
    TopicModelingConfiguration,
)


def _base_config():
    return MarkovAnalysisConfiguration.model_construct(
        schema_version=1,
        text_source=markov.MarkovAnalysisTextSourceConfig.model_construct(),
        segmentation=MarkovAnalysisSegmentationConfig.model_construct(),
        observations=MarkovAnalysisObservationsConfig.model_construct(
            encoder=MarkovAnalysisObservationsEncoder.TFIDF
        ),
        model=MarkovAnalysisModelConfig.model_construct(family=MarkovAnalysisModelFamily.GAUSSIAN),
        topic_modeling=MarkovAnalysisTopicModelingConfig.model_construct(
            enabled=True, configuration=TopicModelingConfiguration.model_construct()
        ),
        llm_observations=MarkovAnalysisLlmObservationsConfig.model_construct(),
    )


def test_apply_topic_modeling_requires_non_boundary_segments():
    config = _base_config()
    boundary = markov.MarkovAnalysisObservation(item_id="i", segment_index=1, segment_text="START")
    with pytest.raises(ValueError):
        markov._apply_topic_modeling(observations=[boundary], config=config, artifacts_dir=None)


def test_encode_observations_categorical_requires_labels():
    config = _base_config()
    config.model = MarkovAnalysisModelConfig.model_construct(family=MarkovAnalysisModelFamily.CATEGORICAL)
    config.observations.categorical_source = "llm_label"
    observations = [
        markov.MarkovAnalysisObservation(item_id="i", segment_index=1, segment_text="a", llm_label="x"),
        markov.MarkovAnalysisObservation(item_id="i", segment_index=2, segment_text="b", llm_label="y"),
    ]
    encoded, lengths = markov._encode_observations(observations=observations, config=config)
    assert encoded == [0, 1]
    assert lengths == [2]


def test_encode_observations_embedded_requires_embeddings():
    config = _base_config()
    config.model = MarkovAnalysisModelConfig.model_construct(family=MarkovAnalysisModelFamily.GAUSSIAN)
    config.observations.encoder = MarkovAnalysisObservationsEncoder.EMBEDDING
    obs = markov.MarkovAnalysisObservation(item_id="i", segment_index=1, segment_text="a", embedding=[0.1, 0.2])
    encoded, lengths = markov._encode_observations(observations=[obs], config=config)
    assert lengths == [1]
    assert encoded == [[0.1, 0.2]]


def test_segment_span_markup_retry_falls_back_to_llm(monkeypatch):
    config = _base_config()
    # enable span markup with prepend_label to hit exception then fallback to _llm_segments
    from biblicus.analysis.models import MarkovAnalysisSpanMarkupSegmentationConfig
    config.segmentation.span_markup = MarkovAnalysisSpanMarkupSegmentationConfig.model_construct(
        prompt_template="tpl",
        client={"provider": "mock", "model": "demo"},
        prepend_label=True,
        label_attribute="label",
        chunk_characters=None,
    )
    def fake_apply_text_extract(request):
        raise ValueError("boom")
    monkeypatch.setattr(markov, "apply_text_extract", fake_apply_text_extract)
    class DummySpan(SimpleNamespace):
        pass
    dummy_result = SimpleNamespace(spans=[SimpleNamespace(text="body", attributes={"label": "LBL"})])
    monkeypatch.setattr(markov, "apply_text_annotate", lambda request: dummy_result)
    monkeypatch.setattr(
        markov,
        "_llm_segments",
        lambda *, item_id, text, config: [MarkovAnalysisSegment(item_id=item_id, segment_index=1, text=text)],
    )
    segments = markov._span_markup_segments(
        item_id="i",
        text="hello world",
        config=config,
    )
    assert segments



import pytest

from biblicus.analysis.models import (
    MarkovAnalysisSpanMarkupSegmentationConfig,
    LlmClientConfig,
)


def test_span_markup_chunk_overlap_requires_chunk_size():
    with pytest.raises(ValueError):
        MarkovAnalysisSpanMarkupSegmentationConfig(
            client=LlmClientConfig(provider="mock"),
            prompt_template="tmpl",
            chunk_overlap_characters=10,
        )


def test_span_markup_chunk_overlap_must_be_smaller():
    with pytest.raises(ValueError):
        MarkovAnalysisSpanMarkupSegmentationConfig(
            client=LlmClientConfig(provider="mock"),
            prompt_template="tmpl",
            chunk_characters=5,
            chunk_overlap_characters=6,
        )



"""
PaddleOCR-VL backed optical character recognition extractor plugin.

This extractor uses PaddleOCR-VL, a vision-language model that provides
improved optical character recognition accuracy especially for complex layouts and multilingual text.

The extractor supports both local inference and application programming interface based inference via
the inference backend abstraction.
"""

from __future__ import annotations

from pathlib import Path
from typing import Any, ClassVar, Dict, List, Optional, Tuple

from pydantic import BaseModel, ConfigDict, Field

from ..corpus import Corpus
from ..errors import ExtractionSnapshotFatalError
from ..inference import ApiProvider, InferenceBackendConfig, InferenceBackendMode, resolve_api_key
from ..models import CatalogItem, ExtractedText, ExtractionStageOutput
from .base import TextExtractor


class PaddleOcrVlExtractorConfig(BaseModel):
    """
    Configuration for the PaddleOCR-VL extractor.

    :ivar retriever: Inference retriever configuration for local or application programming interface execution.
    :vartype retriever: InferenceBackendConfig
    :ivar min_confidence: Minimum confidence threshold for including text.
    :vartype min_confidence: float
    :ivar joiner: String used to join recognized text lines.
    :vartype joiner: str
    :ivar use_angle_cls: Whether to use angle classification for rotated text.
    :vartype use_angle_cls: bool
    :ivar lang: Language code for optical character recognition model.
    :vartype lang: str
    """

    model_config = ConfigDict(extra="forbid", populate_by_name=True)

    retriever: InferenceBackendConfig = Field(
        default_factory=InferenceBackendConfig, alias="backend"
    )
    min_confidence: float = Field(default=0.5, ge=0.0, le=1.0)
    joiner: str = Field(default="\n")
    use_angle_cls: bool = Field(default=True)
    lang: str = Field(default="en")


class PaddleOcrVlExtractor(TextExtractor):
    """
    Extractor plugin using PaddleOCR-VL for optical character recognition.

    This extractor handles image media types and returns text with confidence scores.
    It supports both local inference and application programming interface based inference.

    :ivar extractor_id: Extractor identifier.
    :vartype extractor_id: str
    """

    extractor_id = "ocr-paddleocr-vl"

    _model_cache: ClassVar[Dict[Tuple[str, bool], Any]] = {}

    def validate_config(self, config: Dict[str, Any]) -> BaseModel:
        """
        Validate extractor configuration and ensure prerequisites are available.

        :param config: Configuration mapping.
        :type config: dict[str, Any]
        :return: Parsed configuration model.
        :rtype: PaddleOcrVlExtractorConfig
        :raises ExtractionSnapshotFatalError: If required dependencies are missing.
        """
        import json

        parsed_config = {}
        for key, value in config.items():
            if isinstance(value, str) and (value.startswith("{") or value.startswith("[")):
                try:
                    parsed_config[key] = json.loads(value)
                except json.JSONDecodeError:
                    parsed_config[key] = value
            else:
                parsed_config[key] = value

        parsed = PaddleOcrVlExtractorConfig.model_validate(parsed_config)

        if parsed.retriever.mode == InferenceBackendMode.LOCAL:
            try:
                from paddleocr import PaddleOCR  # noqa: F401
            except ImportError as import_error:
                raise ExtractionSnapshotFatalError(
                    "PaddleOCR-VL extractor (local mode) requires paddleocr. "
                    'Install it with pip install "biblicus[paddleocr]".'
                ) from import_error
        else:
            # api_provider is guaranteed to be set by InferenceBackendConfig validator
            api_key = resolve_api_key(
                parsed.retriever.api_provider,
                config_override=parsed.retriever.api_key,
            )
            if api_key is None:
                provider_name = parsed.retriever.api_provider.value.upper()
                raise ExtractionSnapshotFatalError(
                    f"PaddleOCR-VL extractor (API mode) requires an API key for {provider_name}. "
                    f"Set {provider_name}_API_KEY environment variable or configure "
                    f"{parsed.retriever.api_provider.value} in user config."
                )

        return parsed

    def extract_text(
        self,
        *,
        corpus: Corpus,
        item: CatalogItem,
        config: BaseModel,
        previous_extractions: List[ExtractionStageOutput],
    ) -> Optional[ExtractedText]:
        """
        Extract text from an image using PaddleOCR-VL.

        :param corpus: Corpus containing the item bytes.
        :type corpus: Corpus
        :param item: Catalog item being processed.
        :type item: CatalogItem
        :param config: Parsed configuration model.
        :type config: PaddleOcrVlExtractorConfig
        :param previous_extractions: Prior stage outputs for this item.
        :type previous_extractions: list[ExtractionStageOutput]
        :return: Extracted text with confidence, or None for non-image items.
        :rtype: ExtractedText or None
        """
        _ = previous_extractions

        if not item.media_type.startswith("image/"):
            return None

        parsed_config = (
            config
            if isinstance(config, PaddleOcrVlExtractorConfig)
            else PaddleOcrVlExtractorConfig.model_validate(config)
        )

        source_path = corpus.root / item.relpath

        if parsed_config.retriever.mode == InferenceBackendMode.LOCAL:
            text, confidence = self._extract_local(source_path, parsed_config)
        else:
            api_key = resolve_api_key(
                parsed_config.retriever.api_provider,
                config_override=parsed_config.retriever.api_key,
            )
            text, confidence = self._extract_via_api(source_path, parsed_config, api_key)

        return ExtractedText(
            text=text,
            producer_extractor_id=self.extractor_id,
            confidence=confidence,
        )

    def _extract_local(
        self, source_path: Path, config: PaddleOcrVlExtractorConfig
    ) -> Tuple[str, Optional[float]]:
        """
        Perform local inference using PaddleOCR.

        :param source_path: Path to the image file.
        :type source_path: Path
        :param config: Parsed extractor configuration.
        :type config: PaddleOcrVlExtractorConfig
        :return: Tuple of extracted text and average confidence score.
        :rtype: tuple[str, float or None]
        """
        from paddleocr import PaddleOCR

        cache_key = (config.lang, config.use_angle_cls)
        ocr = PaddleOcrVlExtractor._model_cache.get(cache_key)
        if ocr is None:
            ocr = PaddleOCR(
                use_angle_cls=config.use_angle_cls,
                lang=config.lang,
            )
            PaddleOcrVlExtractor._model_cache[cache_key] = ocr
        result = ocr.ocr(str(source_path))

        if result is None or not result:
            return "", None

        lines: list[str] = []
        confidences: list[float] = []

        # Handle new PaddleOCR API (v4+) that returns dict with rec_texts/rec_scores
        for page_result in result:
            if page_result is None:
                continue

            # New API: dict with 'rec_texts' and 'rec_scores'
            if isinstance(page_result, dict):
                rec_texts = page_result.get('rec_texts', [])
                rec_scores = page_result.get('rec_scores', [])

                for text_value, conf_value in zip(rec_texts, rec_scores):
                    if isinstance(conf_value, (int, float)):
                        confidence = float(conf_value)
                        if confidence >= config.min_confidence:
                            if isinstance(text_value, str) and text_value.strip():
                                lines.append(text_value.strip())
                                confidences.append(confidence)

            # Old API: list of [bbox, (text, confidence)]
            elif isinstance(page_result, (list, tuple)):
                for line_result in page_result:
                    if not isinstance(line_result, (list, tuple)) or len(line_result) < 2:
                        continue
                    text_info = line_result[1]
                    if isinstance(text_info, (list, tuple)) and len(text_info) >= 2:
                        text_value = text_info[0]
                        conf_value = text_info[1]
                        if isinstance(conf_value, (int, float)):
                            confidence = float(conf_value)
                            if confidence >= config.min_confidence:
                                if isinstance(text_value, str) and text_value.strip():
                                    lines.append(text_value.strip())
                                    confidences.append(confidence)

        text = config.joiner.join(lines).strip()
        avg_confidence = sum(confidences) / len(confidences) if confidences else None

        return text, avg_confidence

    def _extract_via_api(
        self, source_path: Path, config: PaddleOcrVlExtractorConfig, api_key: Optional[str]
    ) -> Tuple[str, Optional[float]]:
        """
        Perform inference via application programming interface.

        :param source_path: Path to the image file.
        :type source_path: Path
        :param config: Parsed extractor configuration.
        :type config: PaddleOcrVlExtractorConfig
        :param api_key: Application programming interface key for the provider.
        :type api_key: str or None
        :return: Tuple of extracted text and confidence score.
        :rtype: tuple[str, float or None]
        """
        if config.retriever.api_provider == ApiProvider.HUGGINGFACE:
            return self._extract_via_huggingface_api(source_path, config, api_key)
        else:
            return "", None

    def _extract_via_huggingface_api(
        self, source_path: Path, config: PaddleOcrVlExtractorConfig, api_key: Optional[str]
    ) -> Tuple[str, Optional[float]]:
        """
        Perform inference via HuggingFace Inference API.

        :param source_path: Path to the image file.
        :type source_path: Path
        :param config: Parsed extractor configuration.
        :type config: PaddleOcrVlExtractorConfig
        :param api_key: HuggingFace application programming interface key.
        :type api_key: str or None
        :return: Tuple of extracted text and confidence score.
        :rtype: tuple[str, float or None]
        """
        import base64

        import requests

        with open(source_path, "rb") as f:
            image_data = base64.b64encode(f.read()).decode("utf-8")

        headers = {"Authorization": f"Bearer {api_key}"}

        model_id = config.retriever.model_id or "PaddlePaddle/PaddleOCR-VL"
        api_url = f"https://api-inference.huggingface.co/models/{model_id}"
        response = requests.post(
            api_url,
            headers=headers,
            json={"inputs": image_data},
            timeout=60,
        )
        response.raise_for_status()

        result = response.json()
        return self._parse_api_response(result, config)

    def _parse_api_response(
        self, result: Any, config: PaddleOcrVlExtractorConfig
    ) -> Tuple[str, Optional[float]]:
        """
        Parse application programming interface response.

        :param result: Application programming interface response data.
        :type result: Any
        :param config: Parsed extractor configuration.
        :type config: PaddleOcrVlExtractorConfig
        :return: Tuple of extracted text and confidence score.
        :rtype: tuple[str, float or None]
        """
        _ = config
        if isinstance(result, str):
            return result.strip(), None
        if isinstance(result, dict):
            text = result.get("generated_text", "")
            confidence = result.get("confidence")
            if isinstance(confidence, (int, float)):
                return text.strip(), float(confidence)
            return text.strip(), None
        if isinstance(result, list) and result:
            first = result[0]
            if isinstance(first, dict):
                text = first.get("generated_text", "")
                confidence = first.get("confidence")
                if isinstance(confidence, (int, float)):
                    return text.strip(), float(confidence)
                return text.strip(), None
        return "", None



from __future__ import annotations

import io
import json
from pathlib import Path

import yaml
from behave import given, then, when

from biblicus.corpus import Corpus
from biblicus.hook_manager import HookManager
from biblicus.hooks import HookContext, HookPoint, LifecycleHook
from biblicus.sources import load_source


def _corpus_path(context, name: str) -> Path:
    return (context.workdir / name).resolve()


def _resolve_fixture_path(context, filename: str) -> Path:
    """Resolve fixture path, accounting for corpus root if set."""
    candidate = Path(filename)
    if candidate.is_absolute():
        return candidate
    workdir_path = (context.workdir / candidate).resolve()
    if candidate.parts and candidate.parts[0] == ".biblicus":
        return workdir_path
    corpus_root = getattr(context, "last_corpus_root", None)
    if corpus_root is not None:
        if candidate.parts and candidate.parts[0] == corpus_root.name:
            return workdir_path
        return (corpus_root / candidate).resolve()
    return workdir_path


def _data_for_media_type(media_type: str) -> bytes:
    """
    Provide deterministic fixture data for a media type.

    :param media_type: Internet Assigned Numbers Authority media type.
    :type media_type: str
    :return: Fixture bytes.
    :rtype: bytes
    """
    if media_type == "text/markdown":
        return b"hello\n"
    if media_type.startswith("text/"):
        return b"hello"
    return b"\x89PNG\r\n\x1a\n...binary..."


@given('I have an initialized corpus at "{name}"')
def step_have_initialized_corpus(context, name: str) -> None:
    corpus_path = _corpus_path(context, name)
    context.corpus = Corpus.init(corpus_path)
    context.opened_corpus = None


@when('I open the corpus via the Python application programming interface at "{name}"')
def step_open_corpus_python(context, name: str) -> None:
    corpus_path = _corpus_path(context, name)
    context.opened_corpus = Corpus.open(corpus_path)


@then('the corpus uniform resource identifier starts with "{prefix}"')
def step_corpus_uri_startswith(context, prefix: str) -> None:
    assert context.opened_corpus is not None
    assert context.opened_corpus.uri.startswith(prefix), context.opened_corpus.uri


@when(
    "I ingest an item via the Python application programming interface into corpus "
    '"{name}" with filename "{filename}" and media type "{media_type}"'
)
def step_ingest_item_python(context, name: str, filename: str, media_type: str) -> None:
    corpus_path = _corpus_path(context, name)
    corpus = Corpus.open(corpus_path)
    data = _data_for_media_type(media_type)
    res = corpus.ingest_item(
        data,
        filename=filename,
        media_type=media_type,
        tags=["application-programming-interface"],
        title=None,
        source_uri="python-application-programming-interface",
    )
    context.python_ingest = res
    shown = corpus.get_item(res.item_id)
    context.python_item = json.loads(shown.model_dump_json())


@when(
    "I ingest an item via the Python application programming interface into corpus "
    '"{name}" with no filename and media type "{media_type}"'
)
def step_ingest_item_python_no_filename(context, name: str, media_type: str) -> None:
    corpus_path = _corpus_path(context, name)
    corpus = Corpus.open(corpus_path)
    data = _data_for_media_type(media_type)
    res = corpus.ingest_item(
        data,
        filename=None,
        media_type=media_type,
        tags=["application-programming-interface"],
        title=None,
        source_uri="python-application-programming-interface",
    )
    context.python_ingest = res
    shown = corpus.get_item(res.item_id)
    context.python_item = json.loads(shown.model_dump_json())


@when(
    "I ingest an item via the Python application programming interface with metadata foo "
    '"{value}" into corpus "{name}" with filename "{filename}" and media type "{media_type}"'
)
def step_ingest_item_python_with_metadata(
    context, value: str, name: str, filename: str, media_type: str
) -> None:
    corpus_path = _corpus_path(context, name)
    corpus = Corpus.open(corpus_path)
    data = _data_for_media_type(media_type)
    res = corpus.ingest_item(
        data,
        filename=filename,
        media_type=media_type,
        tags=["application-programming-interface"],
        title=None,
        metadata={"foo": value},
        source_uri="python-application-programming-interface",
    )
    context.python_ingest = res
    shown = corpus.get_item(res.item_id)
    context.python_item = json.loads(shown.model_dump_json())


@when(
    "I ingest a markdown item via the Python application programming interface into corpus "
    '"{name}" with front matter tags and extra tags'
)
def step_ingest_markdown_with_weird_tags(context, name: str) -> None:
    corpus_path = _corpus_path(context, name)
    corpus = Corpus.open(corpus_path)
    md = ("---\n" "tags:\n" "  - x\n" '  - ""\n' "  - 1\n" "---\n" "body\n").encode("utf-8")
    res = corpus.ingest_item(
        md,
        filename="note.md",
        media_type="text/markdown",
        tags=[" ", "y"],
        title=None,
        source_uri="python-application-programming-interface",
    )
    context.python_ingest = res
    shown = corpus.get_item(res.item_id)
    context.python_item = json.loads(shown.model_dump_json())


@then("the python ingest result succeeds")
def step_python_ingest_succeeds(context) -> None:
    assert context.python_ingest is not None
    assert context.python_ingest.item_id


@then('the python ingested item has media type "{media_type}"')
def step_python_item_media_type(context, media_type: str) -> None:
    assert context.python_item is not None
    assert context.python_item.get("media_type") == media_type


@then('the python ingested item relpath ends with "{suffix}"')
def step_python_item_relpath_endswith(context, suffix: str) -> None:
    assert context.python_item is not None
    relpath = context.python_item.get("relpath") or ""
    assert relpath.endswith(suffix), relpath


@then('the python ingested item sidecar includes media type "{media_type}"')
def step_python_item_sidecar_media_type(context, media_type: str) -> None:
    assert context.python_item is not None
    relpath = context.python_item.get("relpath")
    assert isinstance(relpath, str) and relpath
    corpus_root = context.corpus.root
    content_path = corpus_root / relpath
    sidecar_path = content_path.with_name(content_path.name + ".biblicus.yml")
    assert sidecar_path.is_file()
    data = yaml.safe_load(sidecar_path.read_text(encoding="utf-8")) or {}
    assert isinstance(data, dict)
    assert data.get("media_type") == media_type


@then('the python ingested item sidecar includes metadata foo "{value}"')
def step_python_item_sidecar_includes_foo(context, value: str) -> None:
    assert context.python_item is not None
    relpath = context.python_item.get("relpath")
    assert isinstance(relpath, str) and relpath
    corpus_root = context.corpus.root
    content_path = corpus_root / relpath
    sidecar_path = content_path.with_name(content_path.name + ".biblicus.yml")
    assert sidecar_path.is_file()
    data = yaml.safe_load(sidecar_path.read_text(encoding="utf-8")) or {}
    assert isinstance(data, dict)
    assert data.get("foo") == value


@then('the python ingested item tags equal "{csv}"')
def step_python_item_tags_equal(context, csv: str) -> None:
    assert context.python_item is not None
    expected = [t.strip() for t in csv.split(",") if t.strip()]
    assert (context.python_item.get("tags") or []) == expected


@when(
    "I ingest an item via the Python application programming interface into corpus "
    '"{name}" with source uniform resource identifier "{source_uri}" and filename "{filename}"'
)
def step_ingest_item_python_with_source_uri(
    context, name: str, source_uri: str, filename: str
) -> None:
    corpus_path = _corpus_path(context, name)
    corpus = Corpus.open(corpus_path)
    res = corpus.ingest_item(
        b"payload",
        filename=filename,
        media_type="text/plain",
        tags=["application-programming-interface"],
        title=None,
        source_uri=source_uri,
    )
    context.python_ingest = res


@then('the corpus "{name}" hook logs do not include "{text}"')
def step_hook_logs_do_not_include(context, name: str, text: str) -> None:
    corpus_path = _corpus_path(context, name)
    log_dir = corpus_path / "metadata" / "hook_logs"
    combined = ""
    if log_dir.is_dir():
        for p in sorted(log_dir.glob("*.jsonl")):
            combined += p.read_text(encoding="utf-8")
    assert text not in combined


@then('the corpus "{name}" hook logs include "{text}"')
def step_hook_logs_include(context, name: str, text: str) -> None:
    corpus_path = _corpus_path(context, name)
    log_dir = corpus_path / "metadata" / "hook_logs"
    combined = ""
    if log_dir.is_dir():
        for p in sorted(log_dir.glob("*.jsonl")):
            combined += p.read_text(encoding="utf-8")
    assert text in combined


@given('I have a file "{filename}" with contents "{contents}"')
def step_have_file_with_contents(context, filename: str, contents: str) -> None:
    path = _resolve_fixture_path(context, filename)
    path.parent.mkdir(parents=True, exist_ok=True)
    path.write_text(contents, encoding="utf-8")


@given('I have a binary file "{filename}" with bytes:')
def step_have_binary_file_with_bytes(context, filename: str) -> None:
    payload = context.text or ""
    hex_bytes = "".join(payload.split())
    path = _resolve_fixture_path(context, filename)
    path.parent.mkdir(parents=True, exist_ok=True)
    path.write_bytes(bytes.fromhex(hex_bytes))


@given('I have a directory "{path}"')
def step_have_directory(context, path: str) -> None:
    target = _resolve_fixture_path(context, path)
    target.mkdir(parents=True, exist_ok=True)


@when('I load the source "{source}"')
def step_load_source(context, source: str) -> None:
    candidate_path = (context.workdir / source).resolve()
    context.loaded_source = load_source(str(candidate_path))


@when('I load the source uniform resource identifier for "{path}"')
def step_load_source_uri(context, path: str) -> None:
    candidate_path = (context.workdir / path).resolve()
    context.loaded_source = load_source(candidate_path.as_uri())


@when('I attempt to load the source uniform resource identifier for "{path}"')
def step_attempt_load_source_uri(context, path: str) -> None:
    candidate_path = (context.workdir / path).resolve()
    try:
        context.loaded_source = load_source(candidate_path.as_uri())
        context.load_source_error = None
    except Exception as exc:
        context.load_source_error = exc


@then('the source payload filename is "{filename}"')
def step_source_payload_filename(context, filename: str) -> None:
    payload = getattr(context, "loaded_source", None)
    assert payload is not None
    assert payload.filename == filename


@then('the source payload media type is "{media_type}"')
def step_source_payload_media_type(context, media_type: str) -> None:
    payload = getattr(context, "loaded_source", None)
    assert payload is not None
    assert payload.media_type == media_type


@then('the source payload source uniform resource identifier starts with "{prefix}"')
def step_source_payload_source_uri_prefix(context, prefix: str) -> None:
    payload = getattr(context, "loaded_source", None)
    assert payload is not None
    assert payload.source_uri.startswith(prefix), payload.source_uri


@then('the source load error includes "{text}"')
def step_source_load_error_includes(context, text: str) -> None:
    err = getattr(context, "load_source_error", None)
    assert err is not None
    assert text in str(err)


@when("I execute a hook manager with a non-Pydantic hook result")
def step_execute_hook_manager_with_bad_hook(context) -> None:
    class NonPydanticHook:
        hook_id = "non-pydantic"
        hook_points = [HookPoint.before_ingest]

        def run(self, hook_context):
            _ = hook_context
            return {"message": "not a model"}

    manager = HookManager(
        corpus_uri="file://example", log_dir=context.workdir / "logs", hooks=[NonPydanticHook()]
    )
    try:
        manager.run_ingest_hooks(
            hook_point=HookPoint.before_ingest,
            filename="x",
            media_type="text/plain",
            title=None,
            tags=[],
            metadata={},
            source_uri="example",
        )
        context.hook_error = None
    except Exception as exc:
        context.hook_error = exc


@then("a hook result error is raised")
def step_hook_result_error_raised(context) -> None:
    err = getattr(context, "hook_error", None)
    assert err is not None
    assert "non-Pydantic" in str(err) or "non-Pydantic result" in str(err)


@when('I attempt stream ingestion of a Markdown item into corpus "{name}"')
def step_attempt_stream_ingest_markdown(context, name: str) -> None:
    corpus_path = _corpus_path(context, name)
    corpus = Corpus.open(corpus_path)
    try:
        with (context.workdir / "tmp.md").open("wb") as handle:
            handle.write(b"---\n---\n")
        with (context.workdir / "tmp.md").open("rb") as handle:
            corpus.ingest_item_stream(
                handle,
                filename="tmp.md",
                media_type="text/markdown",
                source_uri="file://tmp.md",
            )
        context.stream_ingest_error = None
    except Exception as exc:
        context.stream_ingest_error = exc


@then("the python stream ingestion error is raised")
def step_python_stream_ingest_error(context) -> None:
    err = getattr(context, "stream_ingest_error", None)
    assert err is not None
    assert "Stream ingestion is not supported for Markdown" in str(err)


@when('I stream ingest bytes into corpus "{name}" with no filename and media type "{media_type}"')
def step_stream_ingest_no_filename(context, name: str, media_type: str) -> None:
    corpus_path = _corpus_path(context, name)
    corpus = Corpus.open(corpus_path)
    res = corpus.ingest_item_stream(
        io.BytesIO(b"streamed"),
        filename=None,
        media_type=media_type,
        tags=["application-programming-interface"],
        metadata=None,
        source_uri="python-stream",
    )
    context.python_ingest = res
    shown = corpus.get_item(res.item_id)
    context.python_item = json.loads(shown.model_dump_json())


@when(
    'I stream ingest bytes into corpus "{name}" with filename "{filename}" and media type "{media_type}" and metadata foo "{value}"'
)
def step_stream_ingest_with_metadata(
    context, name: str, filename: str, media_type: str, value: str
) -> None:
    corpus_path = _corpus_path(context, name)
    corpus = Corpus.open(corpus_path)
    res = corpus.ingest_item_stream(
        io.BytesIO(b"streamed"),
        filename=filename,
        media_type=media_type,
        tags=["application-programming-interface"],
        metadata={"foo": value},
        source_uri="python-stream",
    )
    context.python_ingest = res
    shown = corpus.get_item(res.item_id)
    context.python_item = json.loads(shown.model_dump_json())


@when("I call the base lifecycle hook run method")
def step_call_base_hook_run(context) -> None:
    try:
        hook_context = HookContext(
            hook_point=HookPoint.before_ingest,
            operation_id="operation",
            corpus_uri="file://corpus",
            created_at="2000-01-01T00:00:00Z",
        )
        LifecycleHook().run(hook_context)
        context.base_hook_error = None
    except Exception as exc:
        context.base_hook_error = exc


@then("a hook not implemented error is raised")
def step_hook_not_implemented_error(context) -> None:
    err = getattr(context, "base_hook_error", None)
    assert err is not None
    assert isinstance(err, NotImplementedError)


@when("I execute a hook manager with a hook that raises an exception")
def step_execute_hook_manager_with_raising_hook(context) -> None:
    class RaisingHook:
        hook_id = "raising"
        hook_points = [HookPoint.before_ingest]

        def run(self, hook_context):
            _ = hook_context
            raise RuntimeError("boom")

    manager = HookManager(
        corpus_uri="file://example", log_dir=context.workdir / "logs", hooks=[RaisingHook()]
    )
    try:
        manager.run_ingest_hooks(
            hook_point=HookPoint.before_ingest,
            filename="x",
            media_type="text/plain",
            title=None,
            tags=[],
            metadata={},
            source_uri="example",
        )
        context.raising_hook_error = None
    except Exception as exc:
        context.raising_hook_error = exc


@then("a hook execution error is raised")
def step_hook_execution_error_raised(context) -> None:
    err = getattr(context, "raising_hook_error", None)
    assert err is not None
    assert "failed" in str(err) and "boom" in str(err)



"""Tests for TopicClusterer.

BERTopic/UMAP/HDBSCAN are optional; these tests exercise the KMeans fallback
path and all pure-numpy helpers so they run without heavy ML dependencies.
"""

from __future__ import annotations

import numpy as np
import pytest

from biblicus.analysis.reinforcement_memory._clusterer import (
    TopicClusterer,
    _cosine_distance,
)

# ---------------------------------------------------------------------------
# _cosine_distance
# ---------------------------------------------------------------------------


def test_cosine_distance_identical_vectors():
    a = np.array([1.0, 0.0, 0.0])
    assert _cosine_distance(a, a) == pytest.approx(0.0)


def test_cosine_distance_orthogonal_vectors():
    a = np.array([1.0, 0.0])
    b = np.array([0.0, 1.0])
    assert _cosine_distance(a, b) == pytest.approx(1.0)


def test_cosine_distance_opposite_vectors():
    a = np.array([1.0, 0.0])
    b = np.array([-1.0, 0.0])
    assert _cosine_distance(a, b) == pytest.approx(2.0)


def test_cosine_distance_zero_vector_returns_one():
    a = np.array([0.0, 0.0, 0.0])
    b = np.array([1.0, 0.0, 0.0])
    assert _cosine_distance(a, b) == 1.0
    assert _cosine_distance(b, a) == 1.0


# ---------------------------------------------------------------------------
# Helpers to build deterministic small embeddings
# ---------------------------------------------------------------------------


def _make_embeddings(n: int, dim: int = 8, seed: int = 42) -> np.ndarray:
    rng = np.random.default_rng(seed)
    return rng.random((n, dim)).astype(np.float32)


def _make_docs(n: int) -> list[str]:
    return [f"document {i} about topic" for i in range(n)]


# ---------------------------------------------------------------------------
# cluster — KMeans path (n < 15)
# ---------------------------------------------------------------------------


def test_cluster_returns_topic_ids_and_version():
    clusterer = TopicClusterer(min_topic_size=2)
    embeddings = _make_embeddings(6)
    docs = _make_docs(6)
    topic_ids, version = clusterer.cluster(embeddings, docs)
    assert len(topic_ids) == 6
    assert isinstance(version, str)
    assert len(version) > 0


def test_cluster_small_dataset_no_negative_one():
    # KMeans fallback should not produce outlier labels (-1)
    clusterer = TopicClusterer()
    embeddings = _make_embeddings(6)
    docs = _make_docs(6)
    topic_ids, _ = clusterer.cluster(embeddings, docs)
    assert -1 not in topic_ids


def test_cluster_single_item_all_zeros():
    clusterer = TopicClusterer()
    embeddings = _make_embeddings(1)
    docs = _make_docs(1)
    topic_ids, _ = clusterer.cluster(embeddings, docs)
    assert list(topic_ids) == [0]


def test_cluster_two_items():
    clusterer = TopicClusterer()
    embeddings = _make_embeddings(2)
    docs = _make_docs(2)
    topic_ids, _ = clusterer.cluster(embeddings, docs)
    assert len(topic_ids) == 2


# ---------------------------------------------------------------------------
# cluster_centroids
# ---------------------------------------------------------------------------


def test_cluster_centroids_empty_before_clustering():
    assert TopicClusterer().cluster_centroids() == {}


def test_cluster_centroids_after_clustering():
    clusterer = TopicClusterer()
    embeddings = _make_embeddings(6)
    clusterer.cluster(embeddings, _make_docs(6))
    centroids = clusterer.cluster_centroids()
    assert len(centroids) > 0
    for cid, vec in centroids.items():
        assert isinstance(cid, int)
        assert cid != -1
        assert vec.shape == (8,)


def test_cluster_centroids_excludes_outliers():
    clusterer = TopicClusterer()
    # Manually inject topics with outliers
    clusterer._topics = np.array([-1, 0, 0, 1, 1])
    clusterer._embeddings = _make_embeddings(5)
    centroids = clusterer.cluster_centroids()
    assert -1 not in centroids


# ---------------------------------------------------------------------------
# cluster_boundaries
# ---------------------------------------------------------------------------


def test_cluster_boundaries_empty_before_clustering():
    assert TopicClusterer().cluster_boundaries() == {}


def test_cluster_boundaries_after_clustering():
    clusterer = TopicClusterer()
    embeddings = _make_embeddings(6)
    clusterer.cluster(embeddings, _make_docs(6))
    boundaries = clusterer.cluster_boundaries()
    for cid, p95 in boundaries.items():
        assert isinstance(p95, float)
        assert p95 >= 0.0


# ---------------------------------------------------------------------------
# get_keywords
# ---------------------------------------------------------------------------


def test_get_keywords_empty_before_clustering():
    assert TopicClusterer().get_keywords(0) == []


def test_get_keywords_single_doc_returns_empty():
    # TF-IDF needs at least 2 docs to produce meaningful keywords
    clusterer = TopicClusterer()
    embeddings = _make_embeddings(2)
    docs = ["hello world", "foo bar"]
    clusterer.cluster(embeddings, docs)
    # With only one member per cluster (k=1 when n<3), single docs → empty
    # (this tests the <2 docs early-return branch)


def test_get_keywords_with_enough_docs():
    clusterer = TopicClusterer()
    # Force a known cluster manually (bypass BERTopic)
    docs = [
        "pricing confusion about billing",
        "billing statement unclear pricing",
        "customer service response slow",
        "slow response customer service delay",
    ]
    embeddings = _make_embeddings(4)
    clusterer._topics = np.array([0, 0, 1, 1])
    clusterer._embeddings = np.asarray(embeddings, dtype=np.float32)
    clusterer._documents = docs
    kw = clusterer.get_keywords(0, n=5)
    assert isinstance(kw, list)
    # Should surface some pricing/billing related terms


def test_get_keywords_unknown_topic_returns_empty():
    clusterer = TopicClusterer()
    clusterer._topics = np.array([0, 0])
    clusterer._embeddings = _make_embeddings(2)
    clusterer._documents = ["hello world", "foo bar"]
    assert clusterer.get_keywords(99) == []


# ---------------------------------------------------------------------------
# get_representative_exemplars
# ---------------------------------------------------------------------------


def test_get_representative_exemplars_empty_before_clustering():
    assert TopicClusterer().get_representative_exemplars(0) == []


def test_get_representative_exemplars_returns_n_closest():
    clusterer = TopicClusterer()
    docs = ["doc a", "doc b", "doc c", "doc d"]
    embeddings = _make_embeddings(4)
    clusterer._topics = np.array([0, 0, 0, 0])
    clusterer._embeddings = np.asarray(embeddings, dtype=np.float32)
    clusterer._documents = docs
    exemplars = clusterer.get_representative_exemplars(0, n=2)
    assert len(exemplars) == 2
    for idx, text in exemplars:
        assert isinstance(idx, int)
        assert text in docs


def test_get_representative_exemplars_unknown_topic():
    clusterer = TopicClusterer()
    clusterer._topics = np.array([0, 0])
    clusterer._embeddings = _make_embeddings(2)
    clusterer._documents = ["a", "b"]
    assert clusterer.get_representative_exemplars(99) == []


# ---------------------------------------------------------------------------
# generate_labels
# ---------------------------------------------------------------------------


def test_generate_labels_default_label():
    clusterer = TopicClusterer()
    clusterer._topics = np.array([0, 0, 1, 1])
    clusterer._embeddings = _make_embeddings(4)
    clusterer._documents = _make_docs(4)
    labels = clusterer.generate_labels()
    assert labels[0] == "Topic 0"
    assert labels[1] == "Topic 1"


def test_generate_labels_uses_generator():
    calls = []

    def my_gen(docs):
        calls.append(docs)
        return "custom label"

    clusterer = TopicClusterer(label_generator=my_gen)
    clusterer._topics = np.array([0, 0])
    clusterer._embeddings = _make_embeddings(2)
    clusterer._documents = ["hello", "world"]
    labels = clusterer.generate_labels()
    assert labels[0] == "custom label"
    assert len(calls) == 1


def test_generate_labels_empty_before_clustering():
    assert TopicClusterer().generate_labels() == {}


# ---------------------------------------------------------------------------
# get_cluster_records
# ---------------------------------------------------------------------------


def test_get_cluster_records_structure():
    clusterer = TopicClusterer()
    clusterer._topics = np.array([0, 0, 1, 1])
    clusterer._embeddings = _make_embeddings(4)
    clusterer._documents = _make_docs(4)
    clusterer._cluster_version = "20240101-120000"
    records = clusterer.get_cluster_records()
    assert len(records) == 2
    for rec in records:
        assert "doc_id" in rec
        assert "embedding" in rec
        assert "label" in rec
        assert "member_count" in rec
        assert rec["cluster_version"] == "20240101-120000"
        assert isinstance(rec["embedding"], list)


def test_get_cluster_records_empty_before_clustering():
    assert TopicClusterer().get_cluster_records() == []



from pathlib import Path
from types import SimpleNamespace

import pytest

from biblicus import pipelines
from biblicus.models import CorpusCatalog
from biblicus.corpus import Corpus


def test_normalize_extraction_configuration_validates_max_workers():
    with pytest.raises(ValueError):
        pipelines._normalize_extraction_configuration({"max_workers": 0})
    with pytest.raises(ValueError):
        pipelines._normalize_extraction_configuration({"max_workers": True})
    extractor_id, config, workers = pipelines._normalize_extraction_configuration(
        {"extractor_id": "pass-through-text", "configuration": {"foo": "bar"}}
    )
    assert extractor_id == "pipeline"
    assert config["stages"][0]["extractor_id"] == "pass-through-text"
    assert workers is None


def test_run_retrieval_executes_plan(monkeypatch, tmp_path):
    corpus = Corpus(tmp_path)
    corpus.meta_dir.mkdir(parents=True, exist_ok=True)
    catalog = CorpusCatalog(
        schema_version=2,
        generated_at="2024-01-01T00:00:00Z",
        corpus_uri=tmp_path.as_uri(),
        raw_dir="raw",
        items={},
        order=[],
    )
    corpus._write_catalog(catalog)

    called = {"execute": False, "snapshot": False}

    class DummyPlan:
        status = "pending"
        root = SimpleNamespace(reason=None)

        def execute(self, mode, handler_registry):
            called["execute"] = True

    class DummyRetriever:
        def build_snapshot(self, corpus, configuration_name, configuration):
            called["snapshot"] = True
            return SimpleNamespace(snapshot_id="sid")

    monkeypatch.setattr(pipelines, "load_configuration_view", lambda *a, **k: {})
    monkeypatch.setattr(pipelines, "get_retriever", lambda rid: DummyRetriever())
    monkeypatch.setattr(pipelines, "build_plan_for_index", lambda *a, **k: DummyPlan())
    monkeypatch.setattr(pipelines, "build_default_handler_registry", lambda corpus: {})

    config = SimpleNamespace(retriever="scan", configuration=tmp_path / "retrieval.yml")
    snapshot = pipelines._run_retrieval(corpus, config)
    assert called["execute"] and called["snapshot"]
    assert snapshot.snapshot_id == "sid"


def test_run_analysis_loads_configuration(monkeypatch, tmp_path):
    corpus = Corpus(tmp_path)
    corpus.meta_dir.mkdir(parents=True, exist_ok=True)
    catalog = CorpusCatalog(
        schema_version=2,
        generated_at="2024-01-01T00:00:00Z",
        corpus_uri=tmp_path.as_uri(),
        raw_dir="raw",
        items={},
        order=[],
    )
    corpus._write_catalog(catalog)

    loaded = {}

    monkeypatch.setattr(
        pipelines,
        "load_configuration_view",
        lambda *args, **kwargs: loaded.setdefault("config", {"k": "v"}),
    )

    class DummyBackend:
        def run_analysis(self, corpus, configuration_name, configuration, **kwargs):
            loaded["ran"] = (configuration_name, configuration)

    monkeypatch.setattr(pipelines, "get_analysis_backend", lambda kind: DummyBackend())

    analysis_config = SimpleNamespace(kind="markov", configuration=tmp_path / "analysis.yml")
    pipelines._run_analysis(
        corpus,
        analysis_configs=[analysis_config],
        extraction_snapshot=SimpleNamespace(extractor_id="ext", snapshot_id="snap"),
    )
    assert loaded["ran"][0] == "analysis"
    assert loaded["ran"][1] == {"k": "v"}
